from ..containers import Regions

__author__ = 'saulius'
import pandas as pd

def read_cpgs(cpgs_filename):
    cpgs = pd.read_csv(cpgs_filename, sep='\t')

    cols = list(cpgs.columns)
    cols = map(lambda x : x.strip('#'), cols)

    # Rename cols
    new_cols = []
    for c in cols:
        if c == 'chrom':
            c = 'chromosome'
        elif c == 'chromStart':
            c = 'start'
        elif c == 'chromEnd':
            c = 'end'
        new_cols.append(c)

    cpgs.columns = new_cols

    return Regions(cpgs)


import json
import os
import cPickle as pickle
import dgw
from dgw.cluster import add_path_data
from ..dtw.distance import parametrised_dtw_wrapper
import numpy as np

def load_from_pickle(f):
    if isinstance(f, basestring):
        f = open(f, 'rb')
        close = True
    else:
        close = False
    try:
        return pickle.load(f)
    finally:
        if close:
            f.close()

def strict_load(filename):
    try:
        data = load_from_pickle(filename)
        return data
    except Exception, e:
        raise Exception("Failed to read {0!r}, got {1!r}. "
                        "Make sure that you have all files output from DGW in the same directory as the config file"
        .format(filename, e))

def optional_load(filename):
    try:
        data = load_from_pickle(filename)
        return data
    except Exception:
        return None


class Configuration(object):
    """
    Object that stores the major configuration of DGW console applications
    """
    _ignore_directory = None
    _directory = None
    _prefix = None
    _dtw_kwargs = None
    _prototyping_method = None
    _resolution = None

    DEFAULT_FILENAMES = {
        'pairwise_distances': '{prefix}_pairwise_distances.npy',
        'linkage': '{prefix}_linkage.npy',
        'prototypes': '{prefix}_prototypes.pickle',
        'warping_paths': '{prefix}_warping_paths.pickle',
        'config': '{prefix}_config.dgw',
        'missing_regions': '{prefix}_missing_regions.bed',
        'filtered_regions': '{prefix}_filtered_regions.bed',
        'too_short_regions' : '{prefix}_too_short_regions.bed',
        'too_long_regions' : '{prefix}_too_long_regions.bed',
        'no_poi_regions' : '{prefix}_no_poi_regions.bed',
        'dataset': '{prefix}_dataset.pd',
        'raw_dataset': '{prefix}_dataset_raw.pd',
        'regions': '{prefix}_regions.pd',
    }

    FILENAMES = {}

    def __init__(self, args=None, initial_variables=None):
        if args is None and initial_variables is None:
            raise ValueError('Please provide either args or initial_variables')

        if args:
            self._generate_filenames_from_prefix_and_args(args.prefix, args)
            self._dtw_kwargs = self._dtw_kwargs_from_args(args)
            self._prototyping_method = args.prototyping_method
            self._resolution = args.resolution
            self._use_strand_information = args.use_strand_information
        else:
            self.FILENAMES = initial_variables['FILENAMES']
            self._dtw_kwargs = initial_variables['dtw_kwargs']
            self._prototyping_method = initial_variables['prototyping_method']
            self._resolution = initial_variables['resolution']
            self._use_strand_information = initial_variables['use_strand_information']

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, value):
        self._directory = value

    @property
    def resolution(self):
        return self._resolution

    @property
    def use_strand_information(self):
        return self._use_strand_information

    def _generate_filenames_from_prefix_and_args(self, prefix, args):
        """
        Generate default filenames from prefix provided.
        Also parsers the directory

        :param prefix:
        :param args:
        :return:
        """
        self.directory = os.path.dirname(prefix)
        prefix = os.path.basename(prefix)

        self._save_default_filename('config', prefix)

        if not args.blank:
            if args.output_pairwise_distances:
                self._save_default_filename('pairwise_distances', prefix)

            self._save_default_filename('linkage', prefix)
            self._save_default_filename('prototypes', prefix)
            self._save_default_filename('warping_paths', prefix)

        if args.regions:
            self._save_default_filename('regions', prefix)
            self._save_default_filename('missing_regions', prefix)
            self._save_default_filename('filtered_regions', prefix)
            self._save_default_filename('too_short_regions', prefix)

            if args.max_bins:
                self._save_default_filename('too_long_regions', prefix)

        self._save_default_filename('dataset', prefix)
        if args.output_raw_dataset:
            self._save_default_filename('raw_dataset', prefix)

        if args.points_of_interest and args.ignore_no_poi_regions:
            self._save_default_filename('no_poi_regions', prefix)


    def _save_default_filename(self, file_type, prefix):
        self.FILENAMES[file_type] = self.DEFAULT_FILENAMES[file_type].format(prefix=prefix)

    def set_ignore_directory(self, value):
        self._ignore_directory = value

    def _get_filename(self, file_type):
        try:
            return os.path.join(self.directory, self.FILENAMES[file_type])
        except KeyError:
            return None

    @property
    def pairwise_distances_filename(self):
        return self._get_filename('pairwise_distances')

    @property
    def linkage_filename(self):
        return self._get_filename('linkage')

    @property
    def prototypes_filename(self):
        return self._get_filename('prototypes')

    @property
    def warping_paths_filename(self):
        return self._get_filename('warping_paths')

    @property
    def configuration_filename(self):
        return self._get_filename('config')

    @property
    def parsed_regions_filename(self):
        return self._get_filename('regions')

    @property
    def dataset_filename(self):
        return self._get_filename('dataset')

    @property
    def raw_dataset_filename(self):
        return self._get_filename('raw_dataset')

    @property
    def missing_regions_filename(self):
        return self._get_filename('missing_regions')

    @property
    def filtered_regions_filename(self):
        return self._get_filename('filtered_regions')

    @property
    def too_short_regions_filename(self):
        return self._get_filename('too_short_regions')

    @property
    def too_long_regions_filename(self):
        return self._get_filename('too_long_regions')

    @property
    def no_poi_regions_filename(self):
        return self._get_filename('no_poi_regions')

    def _dtw_kwargs_from_args(self, args):
        kw = {'metric': args.metric}
        if args.warping_penalty is not None:
            kw['warping_penalty'] = args.warping_penalty
        if args.slanted_band is not None:
            kw['constraint'] = 'slanted_band'
            kw['k'] = args.slanted_band
        if args.scale:
            kw['scale_first'] = True

        kw['normalise'] = not args.no_length_normalisation
        kw['try_reverse'] = not args.no_reverse
        return kw

    @property
    def dtw_kwargs(self):
        return self._dtw_kwargs

    @property
    def dtw_function(self):
        return parametrised_dtw_wrapper(**self.dtw_kwargs)

    @property
    def prototyping_method(self):
        return self._prototyping_method


    def to_json(self, file):
        data = {'FILENAMES': self.FILENAMES,
                'dtw_kwargs': self.dtw_kwargs,
                'prototyping_method': self._prototyping_method,
                'resolution': self._resolution,
                'use_strand_information' : self._use_strand_information}

        return json.dump(data, file)

    @classmethod
    def from_json(cls, file):
        return Configuration(initial_variables=json.load(file))


    @property
    def blank(self):
        return self.linkage_filename is None

    def load_regions(self):

        return optional_load(self.parsed_regions_filename)

    def load_dataset(self):
        return strict_load(self.dataset_filename)

    def load_linkage(self):
        try:
            linkage = np.load(self.linkage_filename)
        except Exception, e:
            raise IOError('Error reading linkage file {0!r}, got {1!r}', format(linkage, e))
        return linkage

    def load_prototypes(self):
        return strict_load(self.prototypes_filename)

    def load_warping_paths(self):
        return strict_load(self.warping_paths_filename)

    def create_hierarchical_clustering_object(self, dataset=None, regions=None):
        if self.blank:
            raise Exception('Cannot create HierarchicalClustering object from a blank runk')

        if dataset is None:
            dataset = self.load_dataset()

        if regions is None:
            regions = self.load_regions()
        linkage = self.load_linkage()
        prototypes = self.load_prototypes()
        dtw_function = self.dtw_function
        prototyping_method = self.prototyping_method
        warping_paths = self.load_warping_paths()

        hc = dgw.cluster.analysis.HierarchicalClustering(dataset, regions, linkage_matrix=linkage, prototypes=prototypes,
                                                         dtw_function=dtw_function,
                                                         prototyping_method=prototyping_method)
        add_path_data(hc.tree_nodes_list, hc.num_obs, warping_paths)

        return hc





def load_configuration_from_file(configuration_file):
    """
    Loads configuration object from configuration file
    :param configuration_file:
    :return:
    """
    if isinstance(configuration_file, basestring):
        configuration_file = open(configuration_file, 'r')
        should_close = True
    else:
        should_close = False

    try:
        configuration = Configuration.from_json(configuration_file)
    except Exception, e:
        raise Exception('Error opening configuration file provided: {0!r}'.format(e))
    finally:
        if should_close:
            configuration_file.close()

    configuration.directory = os.path.dirname(configuration_file.name)
    return configuration


from StringIO import StringIO
from logging import debug
import pandas as pd
from ..containers import Regions

__author__ = 'saulius'
BED_COLUMNS = ['chromosome', 'start', 'end', 'name', 'score',
               'strand', 'thick_start', 'thick_end', 'item_rgb', 'block_count', 'block_sizes', 'block_starts']

def read_bed(bed_file):
    """
    Parses the bed file specified into `pd.DataFrame`
    :param bed_file:
    :return:
    :rtype: `pd.DataFrame`
    """
    f = open(bed_file, 'r')
    try:
        s = StringIO()
        # Filter out all lines that do not start with "chr" as BED files are allowed to contain some junk
        for line in f:
            if line.startswith('chr'):
                s.write(line)
        s.seek(0)
        regions = pd.read_csv(s, sep="\t", header=None)
    finally:
        f.close()
        s.close()
    regions.columns = BED_COLUMNS[:len(regions.columns)]

    if len(regions.name) != len(regions.name.drop_duplicates()):
        raise Exception('Input BED file {0!r} contains duplicate values in name column. '
                        'Please ensure the names of the regions are unique'.format(bed_file))

    if 'name' in regions.columns:
        regions = regions.set_index('name')


    return Regions(regions)

def write_bed(regions, writable_file, **track_kwargs):
    import csv

    need_closing = False
    if not isinstance(writable_file, file):
        writable_file = open(writable_file, 'w')
        need_closing = True

    if track_kwargs:
        kwarg_strings = {"{0}=\"{1}\"".format(key, value) for key,value in track_kwargs.iteritems()}
        track_header = "track {0}\n".format(' '.join(kwarg_strings))
        writable_file.write(track_header)

    writer = csv.writer(writable_file, delimiter='\t', quotechar='"')
    for ix, data in regions.iterrows():

        row = [data['chromosome'], int(data['start']), int(data['end']), ix]

        # Score is the fifth
        if 'score' in data:
            row.append(data['score'])

        if 'strand' in data:
            if not 'score' in data:
                row.append(0)  # We do not known the score, so make up a value

            row.append(data['strand'])

        writer.writerow(row)

    if need_closing:
        writable_file.close()



import unittest
from numpy.testing import assert_array_equal
import pandas as pd
from dgw.data.containers import Regions
import numpy as np

class TestRegionsClipToResolution(unittest.TestCase):

    def test_clipping_when_only_one_bin_present(self):
        regions = Regions(pd.DataFrame( {'chromosome' : ['chr1', 'chr10'], 'start' : [100, 200], 'end' : [117, 220]} ))

        clipped_df = regions.clip_to_resolution(20)

        self.assertEquals('chr1', clipped_df.ix[0]['chromosome'])
        self.assertEquals(99, clipped_df.ix[0]['start'])
        self.assertEquals(119, clipped_df.ix[0]['end'])

        self.assertEquals('chr10', clipped_df.ix[1]['chromosome'])
        self.assertEquals(200, clipped_df.ix[1]['start'])
        self.assertEquals(220, clipped_df.ix[1]['end'])

    def test_clipping_always_greater_or_equal_than_0(self):
        regions = Regions(pd.DataFrame({'chromosome' : ['chr1'], 'start': [5], 'end': [7]}))

        clipped_df = regions.clip_to_resolution(20)

        self.assertEquals('chr1', clipped_df.ix[0]['chromosome'])
        self.assertEquals(0, clipped_df.ix[0]['start'])
        self.assertEquals(20, clipped_df.ix[0]['end'])

    def test_clipping_multiple_bins(self):
        regions = Regions(pd.DataFrame( {'chromosome' : ['chr1', 'chr10'], 'start' : [100, 200], 'end' : [117, 220]} ))

        clipped_df = regions.clip_to_resolution(5)

        self.assertEquals('chr1', clipped_df.ix[0]['chromosome'])
        self.assertEquals(99, clipped_df.ix[0]['start'])
        self.assertEquals(119, clipped_df.ix[0]['end'])

        self.assertEquals('chr10', clipped_df.ix[1]['chromosome'])
        self.assertEquals(200, clipped_df.ix[1]['start'])
        self.assertEquals(220, clipped_df.ix[1]['end'])

    def test_clipping_res_1(self):

        regions = Regions(pd.DataFrame( {'chromosome' : ['chr1', 'chr10'], 'start' : [100, 200], 'end' : [117, 220]} ))

        clipped_df = regions.clip_to_resolution(1)

        self.assertEquals('chr1', clipped_df.ix[0]['chromosome'])
        self.assertEquals(100, clipped_df.ix[0]['start'])
        self.assertEquals(117, clipped_df.ix[0]['end'])

        self.assertEquals('chr10', clipped_df.ix[1]['chromosome'])
        self.assertEquals(200, clipped_df.ix[1]['start'])
        self.assertEquals(220, clipped_df.ix[1]['end'])

    def test_clipping_keeps_the_same_class(self):

        regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [100, 200], 'end': [117, 220]}))
        clipped_regions = regions.clip_to_resolution(5)
        self.assertTrue(isinstance(clipped_regions, Regions))

class TestRegionsAsBinsOf(unittest.TestCase):

    def test_raises_value_error_on_mismatch(self):
        parent_regions = Regions(pd.DataFrame({'chromosome': ['chr1'], 'start': [100], 'end':[200]}))

        # Wrong chromosome
        poi_regions = Regions(pd.DataFrame({'chromosome': ['chr2'], 'start': [120], 'end':[180]}))
        self.assertRaises(ValueError, poi_regions.as_bins_of, parent_regions)

        # No overlap 1
        poi_regions2 = Regions(pd.DataFrame({'chromosome': ['chr1'], 'start': [20], 'end':[80]}))
        self.assertRaises(ValueError, poi_regions2.as_bins_of, parent_regions)

        # No overlap 2
        poi_regions3 = Regions(pd.DataFrame({'chromosome': ['chr1'], 'start': [220], 'end':[230]}))
        self.assertRaises(ValueError, poi_regions3.as_bins_of, parent_regions)

    def assert_numpy_dicts_equal(self, correct_dict, other_dict):

        self.assertEqual(correct_dict.keys(), other_dict.keys())

        for key in correct_dict:
            c = correct_dict[key]
            o = other_dict[key]

            assert_array_equal(c, o)

    def test_bins_calculated_correctly(self):
        parent_regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [100, 200], 'end': [160, 220]}))

        # Edges
        edge_regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [100, 215], 'end': [110, 220]}))
        correct_ans = {0: np.array([0, 1]), 1: np.array([3])}
        self.assert_numpy_dicts_equal(correct_ans, edge_regions.as_bins_of(parent_regions, resolution=5))

        # Middle
        mid_regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [140, 208], 'end': [145, 212]}))
        correct_ans = {0: np.array([8]), 1: np.array([1, 2])}
        self.assert_numpy_dicts_equal(correct_ans, mid_regions.as_bins_of(parent_regions, resolution=5))

        # Middle, without one index
        mid_regions2 = Regions(pd.DataFrame({'chromosome': ['chr10'], 'start': [208], 'end': [212]}, index=[1]))
        correct_ans = {1: np.array([1, 2])}
        self.assert_numpy_dicts_equal(correct_ans, mid_regions2.as_bins_of(parent_regions, resolution=5))


        length_one_regions = Regions(pd.DataFrame({'chromosome' : ['chr1'], 'start': [140], 'end': [141]}))
        correct_ans = {0: np.array([8])}
        self.assert_numpy_dicts_equal(correct_ans, length_one_regions.as_bins_of(parent_regions, resolution=5))

    def test_named_indices(self):

        parent_regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [100, 200], 'end': [160, 220]},
                                 index=['foo', 'bar']))

        # Edges
        edge_regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [100, 215], 'end': [110, 220]},
                                            index=['foo', 'bar']))
        correct_ans = {'foo': np.array([0, 1]), 'bar': np.array([3])}
        self.assert_numpy_dicts_equal(correct_ans, edge_regions.as_bins_of(parent_regions, resolution=5))


        # Middle
        mid_regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [140, 208], 'end': [145, 212]},
                                           index=['foo', 'bar']))
        correct_ans = {'foo': np.array([8]), 'bar': np.array([1, 2])}
        self.assert_numpy_dicts_equal(correct_ans, mid_regions.as_bins_of(parent_regions, resolution=5))


        # Middle, without one index
        mid_regions2 = Regions(pd.DataFrame({'chromosome': ['chr10'], 'start': [208], 'end': [212]}, index=['bar']))
        correct_ans = {'bar': np.array([1, 2])}
        self.assert_numpy_dicts_equal(correct_ans, mid_regions2.as_bins_of(parent_regions, resolution=5))


        length_one_regions = Regions(pd.DataFrame({'chromosome': ['chr1'], 'start': [140], 'end': [141]}, index=['foo']))
        correct_ans = {'foo': np.array([8])}
        self.assert_numpy_dicts_equal(correct_ans, length_one_regions.as_bins_of(parent_regions, resolution=5))

        regions_not_in_parent = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr2'], 'start': [140, 200], 'end': [141, 300]},
                                                     index=['foo', 'baz']))
        correct_ans = {'foo': np.array([8])}
        self.assert_numpy_dicts_equal(correct_ans, regions_not_in_parent.as_bins_of(parent_regions, resolution=5))



class TestCommonOperationsDoNotBreakRegions(unittest.TestCase):
    """
    Pretty much tests whether the Issue #2859 in pandas is fixed.
    That is if DataFrame object started playing nicely with subclasses
    """

    def test_slicing_works(self):
        regions = Regions(pd.DataFrame({'chromosome': ['chr1', 'chr10'], 'start': [100, 200], 'end': [117, 220]}))
        self.assertTrue(isinstance(regions.head(1), Regions))
        self.assertTrue(isinstance(regions[:1], Regions))




import numpy as np
from pandas.core.nanops import nansum, nanmax

class DataFilter(object):
    """
    Abstract class for DataFilter
    """
    def is_valid(self, data_item):
        """
        This is the method that should be over-rided by subclasses

        :param data_item: a raw array of sequences to consider
        :type data_item: np.ndarray
        :return:
        """
        return True

class MinNumberOfReadsFilter(DataFilter):
    _min_number_of_reads = None

    def __init__(self, min_number_of_reads):
        self._min_number_of_reads = min_number_of_reads

    def is_valid(self, data_row):
        return nansum(data_row) >= self._min_number_of_reads

class HighestPileUpFilter(DataFilter):

    _highest_pileup_threshold = None

    def __init__(self, highest_pileup_threshold):
        self._highest_pileup_threshold = highest_pileup_threshold

    def is_valid(self, data_row):
        return nanmax(data_row) >= self._highest_pileup_threshold



from collections import defaultdict
from logging import debug
from math import floor
from matplotlib.ticker import MultipleLocator
import scipy.cluster.hierarchy as hierarchy
import pandas as pd
import numpy as np
from ..data.containers import AlignmentsData
from ..dtw.distance import dtw_std
from ..dtw import transformations, dtw_projection
from ..dtw.parallel import parallel_dtw_paths

def compute_paths(data, dtw_nodes_list, n, n_processes=None, *dtw_args, **dtw_kwargs):

    non_leaf_nodes = dtw_nodes_list[n:]
    paths = parallel_dtw_paths(data, non_leaf_nodes, n_processes=n_processes, *dtw_args, **dtw_kwargs)
    return paths

def _to_dtw_tree(linkage, hierarchical_clustering_object, prototypes, prototyping_function='mean'):
    """
    Converts a hierarchical clustering linkage matrix `linkage` to hierarchy of `DTWClusterNode`s.
    This is a modification of `scipy.cluster.hierarchy.to_tree` function and the code is mostly taken from it.

    :param linkage: linkage matrix to convert to the DTW Tree
    :param hierarchical_clustering_object: hierarchical clustering object to work with
    :param prototyping_function: "reduce" function for prototype calculation, or "mean" to simply use data mean
    """

    # Validation
    linkage = np.asarray(linkage, order='c')
    hierarchy.is_valid_linkage(linkage, throw=True, name='Z')

    data = hierarchical_clustering_object.data
    labels = data.items
    values = data.ix

    n = linkage.shape[0] + 1

    # Create a list full of None's to store the node objects
    d = [None] * (n * 2 - 1)

    # Create the nodes corresponding to the n original objects.
    for i in xrange(0, n):
        index = labels[i]
        d[i] = DTWClusterNode(id=index, hierarchical_clustering_object=hierarchical_clustering_object,
                              prototype=values[index])

    nd = None

    for i in xrange(0, n - 1):
        fi = int(linkage[i, 0])
        fj = int(linkage[i, 1])

        assert(fi <= i + n)
        assert(fj <= i + n)

        id = i + n
        left = d[fi]
        right = d[fj]
        dist = linkage[i, 2]

        if prototypes:
            prototype = prototypes[id]

            nd = DTWClusterNode(id=id, hierarchical_clustering_object=hierarchical_clustering_object,
                                prototype=prototype,
                                left=left, right=right,
                                dist=linkage[i, 2])

        elif callable(prototyping_function):
            prototype = prototyping_function(left.prototype.values, right.prototype.values, left.count, right.count)

            nd = DTWClusterNode(id=id, hierarchical_clustering_object=hierarchical_clustering_object,
                                prototype=prototype,
                                left=left, right=right,
                                dist=linkage[i, 2])

        elif prototyping_function == 'mean':
            nd = DTWClusterNode(id=id, hierarchical_clustering_object=hierarchical_clustering_object,
                                prototype=None,
                                left=left, right=right,
                                dist=linkage[i, 2])

            # A bit hacky, but does job. Doing this as to get to use nd.data
            nd._prototype = nd.data.mean()

        assert(linkage[i, 3] == nd.count)
        d[n + i] = nd

    return nd, d

def add_path_data(dtw_nodes, n, paths):
    """
    Adds precomputed path data to dtw_nodes
    :param data:
    :param dtw_nodes:
    :param n:
    :param paths:
    :return:
    """
    # Loop through non-leaf nodes
    for node in dtw_nodes[n:]:
        node.warping_paths = paths[node.id]

# Inheritting from object explicitly as hierarchy.ClusterNode is not doing this
class DTWClusterNode(object, hierarchy.ClusterNode):
    _prototype = None
    _index = None
    _hierarchical_clustering_object = None
    _projected_data = None
    _warping_paths = None
    _tracked_points = None
    _tracked_points_histogram = None
    _points_of_interest_histogram = None

    def __init__(self, hierarchical_clustering_object, id, prototype, left=None, right=None, dist=0, count=1):
        hierarchy.ClusterNode.__init__(self, id, left=left, right=right, dist=dist, count=count)
        if not isinstance(prototype, pd.DataFrame):
            prototype = pd.DataFrame(prototype, columns=hierarchical_clustering_object.data.dataset_axis)
        self._prototype = prototype
        self._hierarchical_clustering_object = hierarchical_clustering_object
        self._index = pd.Index(self.__get_item_ids())

        # Assume no points of interest. TODO: Add  way to specify those
        self._points_of_interest = {}

    def __get_item_ids(self):
        """
        Gets the ids of all the leaf nodes that are in the tree below
        :return:
        """
        if self.is_leaf():
            return pd.Index([self.id])
        else:
            return self.get_left().index | self.get_right().index

    @property
    def prototype(self):
        return self._prototype

    @property
    def index(self):
        return self._index

    def reindex(self, new_index):
        # Make sure they are the same elements just in different order
        assert(len(new_index & self._index) == len(self._index) == len(new_index))
        self._index = new_index
        return self._index

    @property
    def data(self):
        # Don't store the data inside
        return self._hierarchical_clustering_object.data.ix[self.index]

    @property
    def projected_data(self):
        # It is infeasible to calculate projected data for all the nodes beforehand
        self.ensure_projections_are_calculated()
        return self._projected_data

    def ensure_projections_are_calculated(self):
        if not self._projected_data:
            self._projected_data = self.__project_items_onto_prototype()

    def _calculate_histogram(self, points_of_interest, number_of_bins, lengths=None):

        histogram = defaultdict(lambda: np.zeros(number_of_bins))
        for ix, poi in points_of_interest.iteritems():
            if lengths is not None:
                scaling_ratio = float(number_of_bins) / lengths[ix]
            else:
                scaling_ratio = 1

            for poi_name, points in poi.iteritems():
                current_histogram = histogram[poi_name]
                for point in set(points):
                    min_rescaled = int(floor(point * scaling_ratio))
                    max_rescaled = int(floor((point + 1) * scaling_ratio))
                    for rescaled_point in xrange(min_rescaled, max_rescaled):
                        current_histogram[rescaled_point] += 1

        return pd.DataFrame(histogram)

    @property
    def tracked_points_histogram(self):
        if self._tracked_points_histogram is None:
            self.ensure_points_of_interest_are_tracked_down()
            self._tracked_points_histogram = self._calculate_histogram(self._tracked_points, len(self.prototype))
        return self._tracked_points_histogram

    @property
    def points_of_interest_histogram(self):
        if self._points_of_interest_histogram is None:
            self._points_of_interest_histogram = self._calculate_histogram(self.points_of_interest,
                                                                           self.data.number_of_columns,
                                                                           self.data.lengths)
        return self._points_of_interest_histogram

    def ensure_points_of_interest_are_tracked_down(self):
        if self._tracked_points is None:
            self._tracked_points = self.__track_points_of_interest()


    @property
    def warping_paths(self):
        if not self._warping_paths:
            raise Exception('Warping paths not computed')

        return self._warping_paths

    @warping_paths.setter
    def warping_paths(self, values):
        self._warping_paths = values

    @property
    def regions(self):
        """
        :rtype: Regions
        """
        if self._hierarchical_clustering_object.regions is None:
            return None
        else:
            return self._hierarchical_clustering_object.regions.ix[self.index]

    def __compute_dtw_warping_paths(self):
        data = self.data
        dtw_function = self._hierarchical_clustering_object.dtw_function
        prototype = self.prototype

        paths = {}
        for ix in data.items:
            _, _, path = dtw_function(data.ix[ix], prototype, dist_only=False)
            # Reduce the bit size of the path arrays to 16 bits
            # DTW would be too slow to use anyway if we had more than 2**16-1 items in it
            # Feel free to update this if it is not the case.
            #path = (np.asarray(path[0], dtype=np.int16), np.asarray(path[1], dtype=np.int16))

            paths[ix] = path

        return paths

    def save_as_encode_track(self, filename, track_name=None, track_description=None):
        if self.regions is None:
            raise Exception('Cannot save {0!r} as region locations are not specified'.format(self))

        if track_name:
            track_kwargs = {'name': track_name}
            if track_description:
                track_kwargs['description'] = track_description
        else:
            track_kwargs = {}
        self.regions.to_bed(filename, **track_kwargs)

    def save_as_list_of_indices(self, filename):
            index = self.data.items
            f = open(filename, 'w')
            for ix in index:
                f.write('{0}\n'.format(ix))
            f.close()



    def __project_items_onto_prototype(self):

        if self.is_leaf():
            # there is no projection, really, return itself
            return self.data
        else:
            data = self.data
            prototype = self.prototype

            warping_paths = self.warping_paths

            columns = data.dataset_axis

            projections = {}
            for ix in data.items:
                item = data.ix[ix]

                projection = dtw_projection(item, prototype, path=warping_paths[ix])
                df = pd.DataFrame(projection, index=range(len(prototype)), columns=columns)
                projections[ix] = df
            panel = pd.Panel(projections)
            panel = panel.ix[self.data.items]
            ad = AlignmentsData(panel)
            return ad

    def __track_points_of_interest(self):

        if self.is_leaf():
            # Nothing to track
            return self.points_of_interest
        else:
            points_of_interest = self.points_of_interest
            warping_paths = self.warping_paths
            tracked_points = defaultdict(lambda: {})
            for ix, pois in points_of_interest.iteritems():
                for j, poi in pois.iteritems():
                    mapped_points = transformations.points_mapped_to(poi, warping_paths[ix])
                    tracked_points[ix][j] = mapped_points

            return tracked_points

    @property
    def points_of_interest(self):
        poi = self.data.points_of_interest
        return poi

    @property
    def tracked_points_of_interest(self):
        self.ensure_points_of_interest_are_tracked_down()
        tracked_points = self._tracked_points
        return tracked_points

    @property
    def n_items(self):
        """
        Returns number of leaves below the current node
        """
        return len(self.index)

    def __len__(self):
        return self.n_items

    def __repr__(self):
        return "<{0} containing {1} items>".format(self.__class__.__name__, self.n_items)


def _reduce_tree(tree, reduce_func, map_function=lambda node: node.id,
                 is_value=lambda x: not isinstance(x, hierarchy.ClusterNode)):
    """
    Performs reduce operation on the binary tree generated by hierarchical clustering.

    Implementation does not use recursion in order not to overflow the stack

    Equivalent to:
    ```
    def reduce(tree, reduce_function, map_function):
    if tree.is_leaf():
        return map_function(tree)
    else:
        left = reduce(tree.get_left(), reduce_function, map_function)
        right = reduce(tree.get_right(), reduce_function, map_function)
        return reduce_function(tree, left, right)
    ```
    :param tree: root node of the tree
    :type tree: `scipy.cluster.hierarchy.ClusterNode`
    :param reduce_func: function that will be performed on reduction. Takes three parameters: node, left branch, right branch
    :param map_function: function that gets the value of a `ClusterNode` defaults to retrieving the node's id
    :param is_value: function that determines whether a (reduced) item or just another node in cluster
    :return: The result of reduction function
    """

    def _add_children_to_stack(node):
        stack.append(node)  # Append self to stack so we can trace it back later
        stack.append(node.get_left())
        stack.append(node.get_right())

    stack = [tree]

    while stack:
        node = stack.pop()
        if not is_value(node):
            if node.is_leaf():
                try:
                    stack.append(map_function(node))
                except Exception, e:
                    print "Got {0!r} while getting value of {1!r}".format(e, node)
                    raise
            else:
                _add_children_to_stack(node)
        else:
            try:
                next_node = stack.pop()
            except IndexError:
                return node

            if not is_value(next_node):
                stack.append(node)
                stack.append(next_node)
            else:
                parent_node = stack.pop()
                assert(not is_value(parent_node))

                try:
                    reduced_value = reduce_func(parent_node, node, next_node)
                except Exception, e:
                    print "Got {0!r} on reduce of {1!r} {2!r}".format(e, node, next_node)
                    raise

                stack.append(reduced_value)

class HierarchicalClustering(object):
    _condensed_distance_matrix = None
    _data = None
    _linkage_matrix = None
    _distance_threshold = None
    __dtw_args = None
    __dtw_kwargs = None
    _regions = None

    __tree = None
    __tree_nodes_list = None


    def __init__(self, data, regions, linkage_matrix, dtw_function=dtw_std, prototypes=None, prototyping_method='psa'):
        """
        Initialises hierarchical clustering analyser.
        Handles linkage calculation, dendrogram plotting and prototype generation.

        Supports three prototyping methods:

            Prioretised Shape Averaging (PSA) (prototyping_method='psa')
                as described in [#Niennattrakul:2009ep], see also `dgw.transformations.sdtw_averaging`
            Standard average of DTW path (prototyping_method='standard')
                averages the DTW paths for each pair of nodes with the same parent.
                The weights are determined by how many sequences were averaged into the node.
                see `dgw.transformations.dtw_path_averaging`
            Unweighted standard average of DTW path (prototyping_method='standard-unweighted')
                similar to the standard method above, but does not bias sequences higher up the tree
            Mean (prototyping_method='mean')
                simply the mean of the data


        .. [#Niennattrakul:2009ep] Vit Niennattrakul and Chotirat Ann Ratanamahatana "Shape averaging under Time Warping",
           2009 6th International Conference on Electrical Engineering/Electronics, Computer,
           Telecommunications and Information Technology (ECTI-CON)

        :param data: a pd.DataFrame object of the data in clusters
        :type data: AlignmentsData
        :param linkage_matrix: linkage_matrix computed by fastcluster.linkage.
        :param dtw_function: DTW calculation function
        :param prototypes: cluster node prototypes (will be computed if None)
        :param prototyping_method: Averaging method either 'psa', 'standard', 'standard-unweighted', 'mean'
        :return:
        """
        if not isinstance(data, AlignmentsData):
            raise ValueError('Data should be instance of {0}'.format(AlignmentsData.__name__))


        self._data = data
        self._regions = regions

        # small negative distances in linkage matrix are sometimes possible due to rounding errors. Change them to 0
        linkage_matrix[:, 2][linkage_matrix[:, 2] < 0] = 0
        self._linkage_matrix = linkage_matrix

        self.__dtw_function = dtw_function

        tree, tree_nodes = self.__dtw_tree_from_linkage(linkage_matrix, prototypes, prototyping_method)
        self.__tree = tree
        self.__tree_nodes_list = tree_nodes

    def extract_prototypes(self):
        prototypes = {}

        for node in self.tree_nodes_list:
            prototypes[node.id] = node.prototype

        return prototypes

    def __dtw_tree_from_linkage(self, linkage, prototypes,  method):
        """
        Computes a prototyped tree from linkage matrix
        :param linkage: linkage matrix
        :param prototypes: possibly precomputed prototypes
        :param method: prototyping method
        :return:
        """



        if method == 'psa':
            averaging_func = lambda x, y, wx, wy: \
                transformations.sdtw_averaging(x, y, wx, wy, dtw_function=self.dtw_function)
        elif method == 'standard':
            averaging_func = lambda x, y, wx, wy: \
                transformations.dtw_path_averaging(x, y, wx, wy, dtw_function=self.dtw_function)
        elif method == 'standard-unweighted':
            averaging_func = lambda x, y, wx, wy: \
                transformations.dtw_path_averaging(x, y, 1, 1, dtw_function=self.dtw_function)
        elif method == 'mean':
            averaging_func = 'mean' # Not a function really, but the code will deal with it
        else:
            raise ValueError('Incorrect method supplied: '
                             'only \'psa\', \'standard\' or \'standard-unweighted\' supported')

        return _to_dtw_tree(linkage, self, prototypes, averaging_func)

    @property
    def data(self):
        """
        Returns the data that is internal to the object
        :rtype: AlignmentsData
        """
        return self._data

    @property
    def regions(self):
        return self._regions

    @property
    def dtw_function(self):
        return self.__dtw_function

    @property
    def condensed_distance_matrix(self):
        return self._condensed_distance_matrix

    @property
    def linkage(self):
        """
        Returns the linkage matrix of the clustering provided
        :return: linkage matrix.
        """
        return self._linkage_matrix


    @property
    def num_obs(self):
        return self.data.number_of_items

    @property
    def dataset_names(self):
        return self.data.dataset_axis

    def as_tree(self):
        return self.__tree

    @property
    def tree_nodes_list(self):
        return self.__tree_nodes_list

    def dendrogram(self, ax=None, no_labels=True, *args, **kwargs):
        """
        Plots the dendrogram for the hieararchical clustering
        :return:
        """
        import matplotlib.pyplot as plt
        linkage = self.linkage
        if ax is None:
            ax = plt.gca()

        color_threshold = kwargs.pop('color_threshold', self._distance_threshold)
        ans = hierarchy.dendrogram(linkage, no_labels=no_labels, color_threshold=color_threshold, *args, **kwargs)

        ax.set_xlabel('Distance')

        return ans



    def _rename_nodes(self, tree):
        """
        Renames the leaf nodes of the cluster tree generated by linkage to match the actual indices of the data.
        :param tree:
        :return:
        """
        stack = [tree]

        while stack:
            node = stack.pop()

            if node.is_leaf():
                node.id = self.data.items[node.id]
            else:
                stack.append(node.get_left())
                stack.append(node.get_right())

    def cut(self, t):
        """
        Cuts the dendrogram at specified threshold t.
        :param t: threshold
           :return:
        """

        self._distance_threshold = t
        root = self.as_tree()

        queue = set([root])

        clusters = set()
        while queue:
            current_node = queue.pop()
            if current_node.dist > t:
                queue.add(current_node.get_right())
                queue.add(current_node.get_left())
            else:
                clusters.add(current_node)
        return ClusterAssignments(self, clusters, t)

    def cut_and_resort(self, cut_threshold, index):
        """
        Cuts the dendrogram based on the color list returned by `scipy.cluster.hierarchy.dedrogram`
        :param cut_threshold: cut threhsold to cut clusters at
        :param index: index of nodes that is already in order
        :return: clusters
        """
        cluster_assignments = self.cut(cut_threshold)


        already_asigned_indices = pd.Index([])
        for cluster in cluster_assignments:
            cluster_index = cluster.index
            sub_index = pd.Index([i for i in index if i in cluster_index])
            cluster.reindex(sub_index)

            if len(already_asigned_indices & sub_index):
                raise Exception("There is some overlap between cluster cuts. There shouldn't be")
            already_asigned_indices =  (already_asigned_indices | sub_index)

        return cluster_assignments


class ClusterAssignments(object):
    _hierarchical_clustering_object = None
    _cluster_roots = None

    _clusters = None
    _cut_depth = None

    def __init__(self, hierarchical_clustering_object, cluster_roots, cut_depth):
        self._hierarchical_clustering_object = hierarchical_clustering_object
        self._cluster_roots = cluster_roots

        clusters = []

        for cluster_root in self._cluster_roots:
            clusters.append(cluster_root)
        self._clusters = clusters
        self._cut_depth = cut_depth

    @property
    def n(self):
        return len(self._cluster_roots)

    def __len__(self):
        return self.n

    @property
    def dataset_names(self):
        return self.hierarchical_clustering_object.dataset_names

    @property
    def cut_depth(self):
        return self._cut_depth

    @property
    def hierarchical_clustering_object(self):
        return self._hierarchical_clustering_object

    @property
    def cluster_sizes(self):
        return pd.Series(map(len, self.clusters))

    def __repr__(self):
        if len(self.clusters) <= 10:
            clusters_repr = '\n'.join(map(repr, self.clusters))
        else:
            clusters_repr = '\n'.join(map(repr, self.clusters[:3]))
            clusters_repr += '\n...\n'
            clusters_repr += '\n'.join(map(repr, self.clusters[-3:]))

        return '<ClusterAssignments n={0}, cut depth: {1}\nClusters: \n{2}>'.format(self.n, self.cut_depth, clusters_repr)

    def flatten(self):
        """
        Flattens the data into a `pd.Series` object that gives a cluster number to every element in original data.
        :return:
        """
        buffer = np.empty(self._hierarchical_clustering_object.num_obs)

        for i, cluster in enumerate(self.clusters):
            queue = set([cluster.root])

            while queue:
                node = queue.pop()

                if node.is_leaf():
                    buffer[node.id] = i + 1
                else:
                    queue.add(node.get_left())
                    queue.add(node.get_right())

        return pd.Series(buffer, index=self.hierarchical_clustering_object.data.index)

    @property
    def clusters(self):
        """
        Returns the data clusters in decreasing number of elements
        :return:
        """
        return self._clusters

    def __iter__(self):
        return iter(self.clusters)

    def __getitem__(self, key):
        return self.clusters[key]



# coding=utf-8
from itertools import imap
from dgw.data.parsers.filters import MinNumberOfReadsFilter
from dgw.dtw import reverse_sequence

__author__ = 'saulius'
from logging import debug
import os

import pandas as pd
import pysam
import numpy as np
from ..containers import AlignmentsData, Regions

def _alignments_that_fall_within_a_region(samfile, chromosome, start, end, extend_to=200):
    """
    Returns raw alignments that fall within the region of interest in samfile.

    :param samfile:
    :param chromosome:
    :param start:
    :param end:
    :param extend_to:
    :return:
    """
    if start < 0:
        raise ValueError('Read start should be greater than 0, {0} given'.format(start))

    # Fetch all alignments from samfile
    if extend_to is None or extend_to == 0:
        read_start = start
        read_end = end
    elif extend_to > 0:
        read_start = max(start - extend_to + 1, 0) # +1 because extended reads should overlap with at least one pixel
        read_end = end + extend_to-1
    else:
        raise ValueError('extend_to should be >= 0')

    try:
        # Converting to list here as otherwise exception wont be caught
        alignments = list(samfile.fetch(chromosome, read_start, read_end))
    except ValueError:
        if extend_to > 0:
            chromosome_length = samfile.lengths[samfile.references.index(chromosome)]
            if end <= chromosome_length < read_end:  # If the exception was caused by extending
                alignments = samfile.fetch(chromosome, read_start, chromosome_length)
            else:
                raise
        else:
            raise

    return alignments

def _read_samfile_region(samfile, chromosome, start, end, resolution=50, extend_to=200):
    """
        Returns piled up read counts for a samfile.
        Chromosome, start, and end all describe a region of the data.
        This should be zero indexed and the last coordinate is not included, following BED format.

        The read count is returned aggregated by number of base pairs equal to resolution parameter.
        That is if resolution is 10 the data will be returned for every 10 base pairs.

        Note that end-start mod resolution should be zero as the function will not know what to do otherwise.

    :param samfile:
    :param chromosome:
    :param start:
    :param end:
    :param resolution:
    :return:
    """
    if (end - start) % resolution != 0:
        raise ValueError('The resolution {0} is not valid for region of length {1}'.format(resolution, end-start))

    # Initialise
    data_len = (end - start) / resolution
    data_buffer = np.zeros(data_len)

    # Fetch all alignments from SamFile
    alignments = _alignments_that_fall_within_a_region(samfile, chromosome, start, end, extend_to=extend_to)
    for alignment in alignments:

        if extend_to is not None and extend_to > 0:
            alignment_start, alignment_end = _extend_read_to(alignment, extend_to)
            if alignment_end < start or alignment_start >= end:
                continue
        else:
            alignment_start = alignment.pos
            alignment_end   = alignment.aend

        start_bin = max(0, (alignment_start-start) / resolution)
        end_bin   = min(data_len -1, (alignment_end-start - 1) / resolution)

        data_buffer[start_bin:end_bin+1] += 1

    return data_buffer

def _extend_read_to(aligned_read, extend_to):

    if aligned_read.alen > extend_to:
        raise ValueError('AlignedRead {0!r} is already longer than {1}'.format(aligned_read, extend_to))

    if extend_to is None or extend_to == 0:
        alignment_start = aligned_read.pos
        alignment_end   = aligned_read.aend
    else:
        alen = aligned_read.alen
        if alen > extend_to:
            raise ValueError('Alignment length, alen={0} greater than extend_to parameter ({1})'.format(alen, extend_to))
        if not aligned_read.is_reverse:
            alignment_start = aligned_read.pos
            alignment_end   = alignment_start + extend_to
        else:
            alignment_start = aligned_read.aend - extend_to
            alignment_end   = aligned_read.aend

    return alignment_start, alignment_end

def read_bam(alignment_filenames, regions, resolution=50, extend_to=200, data_filters=[MinNumberOfReadsFilter(1)],
            output_removed_indices=False, reverse_negative_strand_regions=False):
    """
    Reads provided bam files for the data in the specified regions
    :param alignment_filenames: Filenames of the files to read
    :param regions:         Regions to be explored within these files
    :param resolution:      Resolution at which to read the files
    :param extend_to:       Reads will be extended to extend_to base pairs of length, if not None
    :param data_filters:    Data filters that will be run on data to determine whether to include the value to list or not
    :param output_removed_indices: function outputs both indices in regions that were not found in dataset, and regions removed by filters
    :param reverse_negative_strand_regions: whether to reverse the regions on negative strand or not
    :return: dataset, [regions_not_in_dataset, regions_removed_by_filter] -- the latter two only if output_removed_indices is set

    :rtype: pd.Panel
    """
    # Allow passing either a list of filenames or a single filename
    if isinstance(alignment_filenames, basestring):
        alignment_filenames = [alignment_filenames]

    if len(alignment_filenames) != len(set(alignment_filenames)):
        raise ValueError('Some name in alignments_files provided is duplicated. '
                         'Do you really want to read the file twice?')

    columns = []
    for alignments_file in alignment_filenames:
        name = os.path.basename(alignments_file)
        if name in columns:
            # In case there are files with the same name in differet dirs, use full path
            name = alignments_file
        columns.append(name)


    samfiles = [pysam.Samfile(alignments_file) for alignments_file in alignment_filenames]

    # Get all references that are in all of the files
    references = reduce(lambda x, y: x & y, [set(s.references) for s in samfiles])

    # Get all regions that are in the dataset by chromosome
    # TODO: make sure the regions removed are somewhere accounted for
    dataset_regions = regions[regions.chromosome.isin(references)]

    indices_not_in_dataset = regions.index - dataset_regions.index

    # Make sure the regions play nicely with the resolution
    dataset_regions = dataset_regions.clip_to_resolution(resolution)

    # find the maximum length of a region (we'll use this to initialise numpy arrays)
    max_len = dataset_regions.lengths.max() / resolution

    dataset = {}

    if reverse_negative_strand_regions:
        if not dataset_regions.has_strand_data:
            raise ValueError('reverse_negative_strand_regions is set to true, yet the regions provided have no strand information.')

    for index, region in dataset_regions.iterrows():
        chromosome = region['chromosome']
        start = region['start']
        end = region['end']
        if reverse_negative_strand_regions:
            strand = region['strand']

        data_arr = np.empty((max_len, len(samfiles)))

        for i, samfile in enumerate(samfiles):
            try:
                region_data = _read_samfile_region(samfile,
                                                   chromosome, start, end,
                                                   resolution=resolution,
                                                   extend_to=extend_to)
            except Exception, e:
                raise IOError('Could not read {0}:{1}-{2} from {4}, got: {3!r}'.format(chromosome, start, end, e, alignment_filenames[i]))

            if reverse_negative_strand_regions and strand == '-':
                region_data = reverse_sequence(region_data)

            padding = [np.nan] * (max_len - len(region_data))
            data_arr[:, i] = np.concatenate((region_data, padding))

        valid = all(imap(lambda f: f.is_valid(data_arr), data_filters))
        if not valid:
            continue

        df = pd.DataFrame(data_arr, columns=columns)
        dataset[index] = df

    panel = pd.Panel(dataset)
    data = AlignmentsData(panel)

    filtered_out_indices = dataset_regions.index - data.items

    if output_removed_indices:
        return data, indices_not_in_dataset, filtered_out_indices
    else:
        return data



import unittest
import numpy as np
from dgw.data.parsers.filters import MinNumberOfReadsFilter, HighestPileUpFilter


class TestHighestPileUpFilter(unittest.TestCase):

    def test_all_zeros(self):
        filter = HighestPileUpFilter(1)
        a = np.array([[0, 0], [0, 0], [0, 0], [np.nan, np.nan]])
        self.assertFalse(filter.is_valid(a))

    def test_equality_passes(self):

        filter = HighestPileUpFilter(1)
        a = np.array([1, 1, 1, 1])
        self.assertTrue(filter.is_valid(a))

    def test_max_value_is_considered_only(self):
        filter = HighestPileUpFilter(5)
        a = np.array([1, 2, 3, 4, 17, 0, 0])
        self.assertTrue(filter.is_valid(a))

        b = np.array([1, 2, 3, 4, 0, 0])
        self.assertFalse(filter.is_valid(b))

    def test_any_dimension_is_fine(self):

        filter = HighestPileUpFilter(5)
        a = np.array([[1, 0], [2, 1], [0, 7], [0, 0], [np.nan, np.nan]])
        self.assertTrue(filter.is_valid(a))

class TestMinNumberOfReadsFilter(unittest.TestCase):

    def test_general_cases(self):
        a = np.array([1, 2, 3, np.nan])
        b = np.array([1, 0, 2, np.nan])
        c = np.array([1, 0, 0, np.nan])

        filter = MinNumberOfReadsFilter(3)
        self.assertTrue(filter.is_valid(a))
        self.assertTrue(filter.is_valid(b))  # Equality should pass
        self.assertFalse(filter.is_valid(c))

    def test_dimensions_are_summed(self):
        a = np.array([[1, 2], [2, 1]])
        filter = MinNumberOfReadsFilter(4)

        self.assertTrue(filter.is_valid(a))




from matplotlib.patches import ConnectionPatch
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.ticker import NullFormatter
import numpy as np
from dgw.dtw import reverse_sequence

from distance import dtw_std

def plot_dtw_cost_and_path(cost_matrix, path, ax=None):
    """
        Plots the DTW cost matrix and path.
        Code taken from the mlpy documentation:
        http://mlpy.sourceforge.net/docs/3.5/dtw.html
    :param cost_matrix:
    :param path:
    :return:
    """

    if ax is None:
        ax = plt.gca()

    plot1 = ax.imshow(cost_matrix.T, origin='lower', cmap=cm.gray, aspect='auto', interpolation='nearest')
    plt.colorbar(plot1)
    plot2 = ax.plot(path[0], path[1], 'w')

    xlim = ax.set_xlim((-0.5, cost_matrix.shape[0] - 0.5))
    ylim = ax.set_ylim((-0.5, cost_matrix.shape[1] - 0.5))

def plot_dtw_sequences_dist_cost_and_path(sequence_x, sequence_y, dist, cost_matrix, path):
    """
    Generates a full DTW plot with both the cost matrix, path and both of the sequences

    :param sequence_x:
    :param sequence_y:
    :param dist:
    :param cost_matrix:
    :param path:
    :return:
    """

    sequence_x = np.asarray(sequence_x)
    sequence_y = np.asarray(sequence_y)

    # Based on: http://matplotlib.org/examples/pylab_examples/scatter_hist.html
    figure = plt.figure()
    null_fmt = NullFormatter()         # no labels

    # definitions for the axes
    left, width = 0.1, 0.65
    bottom, height = 0.1, 0.65
    ax_width = 0.2
    bottom_h = left_h = left + width + 0.02

    rect_cost = [left, bottom, width, height]

    rect_sequence_x = [left, bottom_h, width, ax_width]
    rect_sequence_y = [left_h, bottom, ax_width, height]

    text_pos_x = left_h
    text_pos_y = bottom_h

    ax_cost = plt.axes(rect_cost)
    ax_sequence_x = plt.axes(rect_sequence_x)
    ax_sequence_y = plt.axes(rect_sequence_y)

    # Drop labels
    ax_sequence_x.xaxis.set_major_formatter(null_fmt)
    ax_sequence_y.yaxis.set_major_formatter(null_fmt)

    # Plot main plot
    plot_dtw_cost_and_path(cost_matrix, path, ax=ax_cost)

    ax_sequence_x.plot(range(len(sequence_x)), sequence_x)
    ax_sequence_y.plot(sequence_y, range(len(sequence_y)))

    # Set limits based on cost
    ax_sequence_x.set_xlim(ax_cost.get_xlim())
    ax_sequence_y.set_ylim(ax_cost.get_ylim())

    # Set Limits based on seq
    ax_sequence_x.set_ylim(min(sequence_x.min(), sequence_y.min()) - 1,
                           max(sequence_x.max(), sequence_y.max()) + 1)

    ax_sequence_y.set_xlim(min(sequence_x.min(), sequence_y.min()) - 1,
                           max(sequence_x.max(), sequence_y.max()) + 1)

    # Add test showing the value of dist
    plt.figtext(text_pos_x, text_pos_y, 'Distance:\n{0:.5f}'.format(dist), size='medium')

    return figure

def plot_dtw_path(path, ax=None, *args, **kwargs):
    """
    Plots path onto axis supplied by ax (or `plt.gca()` if no such axis given.

    :param path:
    :param ax:
    :param args:
    :param kwargs:
    :return:
    """
    if ax is None:
        ax = plt.gca()

    plot = ax.plot(path[0], path[1], *args, **kwargs)

def visualise_dtw(sequence_x, sequence_y, dtw_function=dtw_std, *args, **kwargs):
    """
    Visualises the DTW distance calculation between sequences x and y using the dtw fucntion provided
    
    :param sequence_x:
    :param sequence_y:
    :param dtw_function:
    :return:
    """
    dist, cost, path = dtw_function(sequence_x, sequence_y, dist_only=False)
    return plot_dtw_sequences_dist_cost_and_path(sequence_x, sequence_y, dist, cost, path, *args, **kwargs)

def visualise_dtw_mappings(sequence_x, sequence_y, dtw_function=dtw_std, columns=None, title=None, sequence_x_label=None,
                           sequence_y_label=None):

    def major_tick_step(ax, axis):
        if axis == 'x':
            ticks = ax.get_xticks()
        else:
            ticks = ax.get_yticks()

        try:
            step = ticks[1] - ticks[0]
        except IndexError:
            step = 0.2

        return step

    def expand_axes(ax):
        x_increment = major_tick_step(ax, 'x') / 8.0

        min_x, max_x = ax.get_xlim()
        ax.set_xlim(min_x - x_increment, max_x + x_increment)

        y_increment = major_tick_step(ax, 'y') / 8.0
        min_y, max_y = ax.get_ylim()
        ax.set_ylim(min_y - y_increment, max_y + y_increment)

    def add_reversed_annotation(ax):
        min_x, max_x = ax.get_xlim()
        min_y, max_y = ax.get_ylim()

        offset_x = major_tick_step(ax, 'x')
        offset_y = major_tick_step(ax, 'y')
        ax.text(min_x + offset_x / 8, max_y - offset_y / 2 - offset_y / 8, '(reversed)')

    dist, cost, path = dtw_function(sequence_x, sequence_y, dist_only=False)

    sequence_x = np.asarray(sequence_x)
    sequence_y = np.asarray(sequence_y)

    try:
        ndim = sequence_x.shape[1]
    except IndexError:
        ndim = 1

    if path[0][0] != 0:
        reversed = True
        sequence_x = reverse_sequence(sequence_x)
        path_x = np.max(path[0]) - path[0]
        path_y = path[1]
        path = (path_x, path_y)
    else:
        reversed = False

    sequence_y_T = np.atleast_2d(sequence_y.T)
    sequence_x_T = np.atleast_2d(sequence_x.T)

    if columns is None and ndim > 1:
        columns = ['Dimension #{0}'.format(i) for i in range(1, ndim+1)]
    elif ndim > 1:
        if len(columns) != ndim:
            raise ValueError('Number of column titles does not match the number of dimensions')

    main_y_axis = None
    xaxes_regular = [None] * ndim
    xaxes_warped = [None] * ndim
    figure = plt.figure()
    figure.subplots_adjust(wspace=0.01, hspace=0.1)
    for i in range(ndim):
        x = sequence_x_T[i]
        print x.shape
        y = sequence_y_T[i]


        ax2 = plt.subplot(2, ndim, ndim + i + 1, sharey=main_y_axis, sharex=xaxes_warped[i])
        ax2.plot(y, color='g')

        if i > 0:
            ax2.yaxis.set_visible(False)

        expand_axes(ax2)

        if not main_y_axis:
            main_y_axis = ax2

        if not xaxes_warped[i]:
            xaxes_warped[i] = ax2

        ax1 = plt.subplot(2, ndim, i + 1, sharey=main_y_axis, sharex=xaxes_regular[i])
        ax1.plot(x, color='b')
        expand_axes(ax1)

        if ndim > 1:
            ax1.set_title(columns[i])

        if i > 0:
            ax1.yaxis.set_visible(False)


        if not xaxes_regular[i]:
            xaxes_regular[i] = ax1
        if reversed:
            add_reversed_annotation(ax1)

        for p_i, p_j in zip(path[0], path[1]):
            xy_a = (p_i, x[p_i])
            xy_b = (p_j, y[p_j])
            con = ConnectionPatch(xyA=xy_a, xyB=xy_b, coordsA="data", coordsB="data",
                                 axesA=ax1, axesB=ax2, arrowstyle="-", shrinkB=2, shrinkA=2, alpha=0.2)

            ax1.add_artist(con)

    if title is not None:
        plt.suptitle(title)

    lines = xaxes_regular[0].get_lines()
    lines.extend(xaxes_warped[0].get_lines())

    if sequence_x_label and sequence_y_label:
        plt.figlegend(lines, (sequence_x_label, sequence_y_label), 'lower center')

    return figure


from dgw.data.parsers import bam as bam_parser

__author__ = 'saulius'
import unittest
from numpy.testing import *
import numpy as np

# -- Stub classes to simulate pysam behaviour ---
class StubAlignedRead(object):
    def __init__(self, pos, alen, is_reverse):
        self.pos = pos
        self.alen = alen
        self.is_reverse = is_reverse

    @property
    def aend(self):
        return self.pos + self.alen

class StubSamfile(object):
    __fetch_response = None
    __references = None
    __lengths = None

    def __init__(self, fetch_response, references, lengths):
        """
        Simulates `Samfile` behaviour.
        Returns `fetch_response` for all responses that satisfy the following criterion:
           `chromosome` is in `references`,
           0 <= `start`, end <= length_of_chromosome
           where length_of_chromosome = lengths[references.index('chromosome')]

        :param fetch_response: the response that will be returned
        :param references: a list of references (chromosomes) in the genome
        :param lengths: of these references
        :return:
        """
        self.__fetch_response = fetch_response
        self.__references = references
        self.__lengths = lengths

    @property
    def references(self):
        return self.__references

    @property
    def lengths(self):
        return self.__lengths

    def fetch(self, chromosome, start, end):
        if chromosome not in self.references:
            raise ValueError('Invalid reference')

        chromosome_length = self.lengths[self.references.index(chromosome)]

        if start < 0:
            raise ValueError('Start out of bounds {0} < 0'.format(start))
        elif end > chromosome_length:
            raise ValueError('End out of bounds {0} > {1}'.format(end, chromosome_length))

        for item in self.__fetch_response:
            if item.pos < end and item.aend >= start:
                yield item

# -- Tests ------------------------------------------------------------

class TestReadExtendFunctions(unittest.TestCase):
    def test_extend_regular_read(self):
        aligned_read = StubAlignedRead(100, 36, False)
        start, end = bam_parser._extend_read_to(aligned_read, 100)

        self.assertEqual(100, start)
        self.assertEqual(200, end)

    def test_extend_reverse_read(self):
        aligned_read = StubAlignedRead(164, 36, True)
        start, end = bam_parser._extend_read_to(aligned_read, 100)

        self.assertEqual(100, start)
        self.assertEqual(200, end)

    def test_invalid_extend_raises_exception(self):
        aligned_read = StubAlignedRead(100, 36, True)

        with self.assertRaises(ValueError):
            bam_parser._extend_read_to(aligned_read, 10) # 0 < 36

class TestReadCountForRegionReading(unittest.TestCase):

    def setUp(self):

        aligned_reads = [ StubAlignedRead(10, 10, True), # A
                          StubAlignedRead(20, 10, True),  # B
                          StubAlignedRead(5, 10, False),   # C
                          StubAlignedRead(25, 10, False)   # D
                          ]

        #01234567890123456789012345678901234
        #..........AAAAAAAAAABBBBBBBBBB.....
        #.....CCCCCCCCCC..........DDDDDDDDDD
        self.samfile = StubSamfile(aligned_reads, ['chr1', 'chr2'], [50, 20])

    def test_non_extended_read_no_binning(self):
        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 3, 40, resolution=1, extend_to=None)
        correct = np.array([0,0,1,1,1,1,1,2,2,2,2,2,1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,1,1,1,1,1,0,0,0,0,0])

        assert_array_equal(correct, peak_data)

    def test_non_extended_read_binning(self):

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 3, 43, resolution=5, extend_to=None)
        correct = np.array([1,2,2,2,2,2,1,0])

        assert_array_equal(correct, peak_data)

    def test_extended_read_no_binning(self):
        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 3, 40, resolution=1, extend_to=15)

        #012|3456789012345678901234567890123456789|
        #...|..aaaaaAAAAAAAAAA....................|
        #...|............bbbbbBBBBBBBBBB..........|
        #...|..CCCCCCCCCCccccc.....DDDDDDDDDDddddd|

        correct = np.array([0,0,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,1,1,1,1,1,2,2,2,2,2,1,1,1,1,1,1,1,1,1,1])
        assert_array_equal(correct, peak_data)


    def test_extended_read_binning(self):
        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 3, 43, resolution=5, extend_to=15)

        #012[34567|89012|34567|89012|34567|89012|34567|89012]
        #...[..aaa|aaAAA|AAAAA|AA...|.....|.....|.....|.....]
        #...[.....|.....|..bbb|bbBBB|BBBBB|BB...|.....|.....]
        #...[..CCC|CCCCC|CCccc|cc...|..DDD|DDDDD|DDddd|dd...]

        correct = np.array([2,2,3,3,2,2,1,1])
        assert_array_equal(correct, peak_data)

    def test_extended_read_no_binning_extended_boundaries(self):

        #0123456789012345[67]8901234567890123456789
        #.....aaaaaAAAAAA[AA]AA....................
        #...............b[bb]bbBBBBBBBBBB..........
        #.....CCCCCCCCCCc[cc]cc.....DDDDDDDDDDddddd

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 16, 18, resolution=1, extend_to=15)
        correct = np.array([3,3])
        assert_array_equal(correct, peak_data)

        #012345678901234567890123456789012345678[9012]
        #.....aaaaaAAAAAAAAAA...................[....]
        #...............bbbbbBBBBBBBBBB.........[....]
        #.....CCCCCCCCCCccccc.....DDDDDDDDDDdddd[d...]

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 39, 43, resolution=1, extend_to=15)
        correct = np.array([1,0,0,0])
        assert_array_equal(correct, peak_data)

        #01234567890123456789012345678901234567890[123]
        #.....aaaaaAAAAAAAAAA.....................[...]
        #...............bbbbbBBBBBBBBBB...........[...]
        #.....CCCCCCCCCCccccc.....DDDDDDDDDDddddd[....]

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 41, 44, resolution=1, extend_to=15)
        correct = np.array([0,0,0])
        assert_array_equal(correct, peak_data)

    def test_extended_read_no_binning_extended_boundaries_edge_case(self):

        #[012345]67890123456789012345678901234567890123
        #[.....a]aaaaAAAAAAAAAA........................
        #[......].........bbbbbBBBBBBBBBB..............
        #[.....C]CCCCCCCCCccccc.....DDDDDDDDDDddddd....

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 0, 6, resolution=1, extend_to=15)
        correct = np.array([0,0,0,0,0,2])
        assert_array_equal(correct, peak_data)

        #[01234]567890123456789012345678901234567890123
        #[.....]aaaaaAAAAAAAAAA........................
        #[.....]..........bbbbbBBBBBBBBBB..............
        #[.....]CCCCCCCCCCccccc.....DDDDDDDDDDddddd....

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 0, 5, resolution=1, extend_to=15)
        correct = np.array([0,0,0,0,0])
        assert_array_equal(correct, peak_data)

    def test_extended_reads_extended_boundaries_binning(self):

        #012345678901234567[8901234|5678901|2345678]90123
        #.....aaaaaAAAAAAAA[AA.....|.......|.......].....
        #...............bbb[bbBBBBB|BBBBB..|.......].....
        #.....CCCCCCCCCCccc[cc.....|DDDDDDD|DDDdddd]d....

        peak_data = bam_parser._read_samfile_region(self.samfile, 'chr1', 18, 39, resolution=7, extend_to=15)
        correct = np.array([3,2,1])
        assert_array_equal(correct, peak_data)

    def test_read_samfile_region_raises_exception_when_read_region_is_bad(self):
        self.assertRaises(ValueError, bam_parser._read_samfile_region, self.samfile, 'chr1', -20, 20, resolution=4, extend_to=15)
        self.assertRaises(ValueError, bam_parser._read_samfile_region, self.samfile, 'chr1', 10, 60, resolution=4, extend_to=15)

if __name__ == '__main__':
    unittest.main()


from math import floor, fabs, ceil
import numpy as np
from dgw.dtw.utilities import _strip_nans


def uniform_scaling_to_length(sequence, desired_length, output_scaling_path=False):
    """
    Uniform scaling procedure, similar to the one provided in [#yankov2007]
    .. [#yankov2007] D Yankov, E Keogh, J Medina, and B Chiu, "Detecting time series motifs under uniform scaling", 2007
    :param sequence:
    :param desired_length:
    :return:
    """
    sequence = _strip_nans(sequence)
    current_len = len(sequence)
    if current_len == 0:
        raise ValueError('Empty sequence cannot be extended')
    elif desired_length == current_len:
        if output_scaling_path:
            return sequence, range(desired_length)
        else:
            return sequence
    elif desired_length < current_len:
        raise ValueError('Desired length is smaller than current length: {0} < {1}'.format(desired_length, current_len))

    scaling_factor = float(current_len) / desired_length

    rescaled_sequence = [sequence[int(floor(i*scaling_factor))] for i in range(desired_length)]

    if output_scaling_path:
        return rescaled_sequence, np.asarray([int(floor(i * scaling_factor)) for i in range(desired_length)])
    else:
        return rescaled_sequence


def uniform_shrinking_to_length(sequence, desired_length):
    EPSILON = 1e-6

    sequence = np.asarray(sequence, dtype=float)
    sequence = _strip_nans(sequence)

    current_length = len(sequence)

    if current_length == 0:
        raise ValueError('Cannot shrink sequence of length 0')
    elif current_length < desired_length:
        raise ValueError('Desired length greater than current length: {0} > {1}'.format(desired_length, current_length))
    elif current_length == desired_length:
        return sequence

    if desired_length <= 0:
        raise ValueError('Invalid length desired: {0}'.format(desired_length))

    # This is essentially how many points in the current sequence will be mapped to a single point in the newone
    shrink_factor = float(current_length) / desired_length

    try:
        ndim = sequence.shape[1]
    except IndexError:
        ndim = 0
    if ndim == 0:
        new_sequence = np.empty(desired_length)
    else:
        new_sequence = np.empty((desired_length, ndim))

    for i in range(desired_length):
        start = i * shrink_factor
        end = (i + 1) * shrink_factor

        s = 0
        d = 0

        left_bound = int(floor(start))

        if fabs(start - left_bound) <= EPSILON:
            left_bound_input = 1
        else:
            left_bound_input = ceil(start) - start

        if left_bound_input >= EPSILON:
            s += sequence[left_bound] * left_bound_input
            d += left_bound_input

        right_bound = int(floor(end))
        right_bound_input = end - floor(end)

        if right_bound_input >= EPSILON:  # Epsilon to prevent rounding errors interfering
            s += sequence[right_bound] * right_bound_input
            d += right_bound_input

        for j in xrange(left_bound + 1, right_bound):
            s += sequence[j]
            d += 1.0

        assert(abs(d - shrink_factor) < 0.000001)

        new_sequence[i] = s / d

    return new_sequence


import unittest
from math import sqrt
from scipy.spatial.distance import cosine
import numpy as np
from numpy.testing import *

from dgw.dtw.distance import dtw_std
from dgw.dtw.utilities import _strip_nans, reverse_sequence


class TestStripNans(unittest.TestCase):

    def test_single_dimension_no_strip_needed(self):
        x = np.array([1,2,3,4,5,6], dtype=float)
        assert_array_equal(x, _strip_nans(x))


    def test_more_dims_no_strip_needed(self):
        x = np.array([[1,2,3,4,5,6], [7,8,9,10,11,12], [13,14,15,16,17,18]], dtype=float).T
        assert_array_equal(x, _strip_nans(x))

    def test_single_dimension_with_strip(self):
        x = np.array([1,2,3,4,5,np.nan, np.nan], dtype=float)
        assert_array_equal(np.array([1,2,3,4,5], dtype=float), _strip_nans(x))

    def test_multi_dimension_with_strip(self):
        x = np.array([[1,2,3,np.nan, np.nan, np.nan], [7,8,9,np.nan, np.nan, np.nan], [13,14,15,np.nan,np.nan,np.nan]], dtype=float).T
        correct = np.array([[1,2,3], [7,8,9], [13, 14, 15]], dtype=float).T
        assert_array_equal(correct, _strip_nans(x))

    def test_multi_dimension_inconsistent_nans(self):
        x = np.array([[1,2,3,4,5,np.nan], [7,8,9,10,11,np.nan], [13,14,15,16,17,18]], dtype=float).T
        self.assertRaises(ValueError, _strip_nans, x)

        x = np.array([[1,2,3,4,5,np.nan], [7,8,9,10,11,np.nan], [13,14,15,16,np.nan,18]], dtype=float).T
        self.assertRaises(ValueError, _strip_nans, x)

class TestReverse(unittest.TestCase):

    def test_single_dimension(self):
        a = np.array([1,2,3,4,5,6])
        assert_array_equal(np.array([6,5,4,3,2,1]), reverse_sequence(a))

    def test_multi_dimension(self):
        a = np.array([[1, 2], [3,4], [5,6]])
        assert_array_equal(np.array([[5,6], [3,4], [1,2]]), reverse_sequence(a))

    def test_single_dimension_with_nans_appended(self):
        a = np.array([1,2,3,4,np.nan, np.nan])
        assert_array_equal(np.array([4, 3, 2, 1, np.nan, np.nan]), reverse_sequence(a))

    def test_multi_dimension_with_nans_appended(self):
        a = np.array([[1,2],[3,4],[np.nan, np.nan]])
        assert_array_equal(np.array([[3,4],[1,2], [np.nan, np.nan]]), reverse_sequence(a))

class TestDTWStd(unittest.TestCase):

    def test_multidimensional_dtw(self):

        a = np.array([[1,2,3, np.nan], [7,8,9,np.nan]]).T
        b = np.array([[10,12,14], [13,15,17]]).T

        # DTW should match the points:
        # (1,7) to (10,13) (distance: sqrt(81 + 36 = 117))
        # (2,8) to (12,15) (distance: sqrt(100 + 49 = 149))
        # (3,9) to (14,17) (distance: sqrt(121 + 64 = 185))

        # Euclidean distance is the one worth testing for, as sqeuclidean will be the same
        # for a.T and b.T as well.
        euclid_distance = sqrt(117) + sqrt(149) + (sqrt(185))
        self.assertAlmostEqual(euclid_distance, dtw_std(a, b, metric='euclidean'))

    def test_slanted_band_dtw(self):

        a = np.array([1, 2, 3, 4, 5, 6, 7])
        b = np.array([1, 1, 1, 1, 4, 4, 4])

        # Slanted band should constrain mapping all elements exactly
        correct_dist = 0 + 1 + 2 ** 2 + 3 ** 2 + 1 + 2 ** 2 + 3 ** 2
        dist = dtw_std(a, b, constraint='slanted_band', k=0)

        self.assertEqual(correct_dist, dist)

    def test_multidimensional_dtw_cosine(self):

        a = np.array([[1,2,3, np.nan], [7,8,9,np.nan]]).T
        b = np.array([[10,12,14], [13,15,17]]).T

        # DTW should match the points:
        # (1,7) to (10,13)
        # (2,8) to (12,15)
        # (3,9) to (14,17)

        #
        cosine_distance = cosine(a[0], b[0]) + cosine(a[1], b[1]) + cosine(a[2], b[2])
        self.assertAlmostEqual(cosine_distance, dtw_std(a, b, metric='cosine'))

    def test_reverse_dtw(self):

        a = np.array([1, 2, 3, np.nan])
        b = np.array([3, 2, 1])

        self.assertEqual(0, dtw_std(a, b, try_reverse=True))

        dist, cost, path = dtw_std(a, b, try_reverse=True, dist_only=False)

        assert_array_equal([2, 1, 0], path[0])
        assert_array_equal([0, 1, 2], path[1])

    def test_multidimensional_dtw_reverse(self):

        a = np.array([[1,2,3, np.nan], [7,8,9,np.nan]]).T
        b = np.array([[10,12,14], [13,15,17]]).T

        #Reverse:
        b = reverse_sequence(b)

        # DTW should match the points:
        # (1,7) to (10,13) (distance: sqrt(81 + 36 = 117))
        # (2,8) to (12,15) (distance: sqrt(100 + 49 = 149))
        # (3,9) to (14,17) (distance: sqrt(121 + 64 = 185))

        # Euclidean distance is the one worth testing for, as sqeuclidean will be the same
        # for a.T and b.T as well.
        euclid_distance = sqrt(117) + sqrt(149) + (sqrt(185))
        self.assertAlmostEqual(euclid_distance, dtw_std(a, b, metric='euclidean', try_reverse=True))







import numpy as np

__author__ = 'saulius'


def _strip_nans(sequence):
    '''
        Strips NaNs that are padded to the right of each peak if they are of unequal length
    :param sequence:
    :return:
    '''
    sequence = np.asarray(sequence)

    try:
        lookup = np.all(np.isnan(sequence), axis=1)
    except ValueError:
        # Will get thrown for one-dimensional arrays
        return sequence[~np.isnan(sequence)]

    sequence = sequence[~lookup]

    if np.any(np.isnan(sequence)):
        raise ValueError('Inconsistent NaNs between dimensions')

    return sequence


def no_nans_len(sequence):
    """
    Returns length of the sequence after removing all the nans from it.

    :param sequence:
    :return:
    """
    return len(_strip_nans(sequence))

def reverse_sequence(sequence):
    no_nans_sequence = _strip_nans(sequence)
    rev = no_nans_sequence[::-1]

    padding = len(sequence) - len(no_nans_sequence)

    try:
        ndim = no_nans_sequence.shape[1]
    except IndexError:
        return np.concatenate((rev, [np.nan] * padding))

    if padding:
        nans = [[np.nan] * ndim] * padding
        return np.concatenate((rev, nans))
    else:
        return rev


from logging import debug
import pandas as pd
import numpy as np
from dgw.data.parsers.pois import map_to_bins
from dgw.dtw import no_nans_len


class AlignmentsDataIndexer(object):
    """
    A wrapper around `_NDFrameIndexer` that would return `AlignmentsData` objects instead of `pd.Panel` objects
    """
    _ndframe_indexer = None
    _alignments_data = None

    def __init__(self, ndframe_indexer, alignments_data):
        self._ndframe_indexer = ndframe_indexer
        self._alignments_data = alignments_data

    def __getitem__(self, key):
        result = self._ndframe_indexer.__getitem__(key)
        if isinstance(result, pd.Panel):
            data = AlignmentsData(result)
            data.points_of_interest = self._alignments_data.points_of_interest
            return data
        else:
            return result

    def __setitem__(self, key, value):
        self._ndframe_indexer.__setitem__(self, key, value)

class RegionsIndexer(object):
    """
    A wrapper around `_NDFrameIndexer` that would return `Regions` objects instead of `pd.DataFrame` objects
    """
    _ndframe_indexer = None
    regions = None

    def __init__(self, ndframe_indexer, regions):
        self._ndframe_indexer = ndframe_indexer
        self.regions = regions

    def __getitem__(self, key):
        result = self._ndframe_indexer.__getitem__(key)
        if isinstance(result, pd.DataFrame):
            data = self.regions.__class__(result)
            return data
        else:
            return result

    def __setitem__(self, key, value):
        self._ndframe_indexer.__setitem__(self, key, value)

class AlignmentsData(object):

    _data = None
    _poi = None
    _scale = None

    def __init__(self, panel, poi=None, scale='raw'):
        """
        Initialises `AlignmentsData` with a `panel` provided.
        The panel is assumed to have data sets on the minor axis
        See `dgw.data.parsers.read_bam` for how to generate this data.

        :param panel: `pd.Panel` object `AlignmentsData` will be initialised on, or `pd.DataFrame` that will be converted
                     to Panel
        :param scale: the scale of data
        :return:
        """
        if isinstance(panel, pd.DataFrame):
            # Create a panel from the DataFrame by giving it a generic name and making sure it is on the minor axis
            self._data = pd.Panel({'Dataset 1'}).transpose(1, 2, 0)
        elif isinstance(panel, pd.Panel):
            self._data = panel
        else:
            raise Exception('Invalid type of data provided for AlignmentsData: {0!r}, expected pd.Panel or pd.Dataframe'
                            .format(type(panel)))

        self._scale = scale

        self.points_of_interest = poi


    def reset_poi(self):
        self._poi = {}

    @property
    def data(self):
        return self._data

    @property
    def points_of_interest(self):
        return self._poi

    @points_of_interest.setter
    def points_of_interest(self, value):
        if value is None:
            self._poi = {}
        else:
            self._poi = {ix: value for ix, value in value.iteritems() if ix in self.items}


    def add_points_of_interest(self, binned_points_of_interest_regions, name):
        for ix, value in binned_points_of_interest_regions.iteritems():
            if ix not in self.items:
                continue

            try:
                self._poi[ix][name] = value
            except KeyError:
                self._poi[ix] = {name: value}


    def drop_no_pois(self):
        common_index = self.items & self.points_of_interest

        return self.ix[common_index]

    #-- Functions that simulate pd.Panel behaviour -------------------------------------------------------------------

    def mean(self, axis='items', skipna=True):
        # Override axis parameter in the pd.Panel mean function
        return self._data.mean(axis=axis, skipna=skipna)

    @property
    def values(self):
        return self.data.values


    @property
    def number_of_datasets(self):
        return len(self.dataset_axis)

    @property
    def number_of_items(self):
        return len(self.items)

    @property
    def number_of_columns(self):
        return len(self.major_axis)

    @property
    def lengths(self):
        a = []
        for _, row in self.data.iteritems():
            a.append(no_nans_len(row.values))

        return pd.Series(a, index=self.items)


    @property
    def dataset_axis(self):
        return self.data.minor_axis

    def dataset_xs(self, *args, **kwargs):
        return self.data.minor_xs(*args, **kwargs)

    @property
    def items(self):
        return self.data.items

    @property
    def ix(self):
        return AlignmentsDataIndexer(self.data.ix, self)

    def __getitem__(self, item):
        return self.data.__getitem__(item)

    def head(self, n=5):
        return self.ix[:n]

    @property
    def major_axis(self):
        return self.data.major_axis


    #-- Additional transformations not visible in default pd.Panel  ----------------------------------------
    def __len__(self):
        return self.number_of_items

    def to_log_scale(self):
        if self._scale == 'log':
            return self

        new_data = (self.data + 2).apply(np.log)  # Adding +2 so we have no zeros in log output
        ad = AlignmentsData(new_data, scale='log')
        ad.points_of_interest = self.points_of_interest
        return ad

    def normalise_bin_heights(self):
        data = {}
        for ix, data_row in self.data.iteritems():
            data[ix] = data_row / data_row.max()

        data = pd.Panel(data)
        return self.__class__(data, poi=self.points_of_interest)

    def plot_heatmap(self, *args, **kwargs):
        """
        Plots heatmap of the data stored in the panel.

        :param args: args to be passed into `visualisation.heatmap.plot`
        :param kwargs: kwargs to be passed into `visualisation.heatmap.plot`
        :return:
        """
        import visualisation.heatmap
        return visualisation.heatmap.plot(self, *args, **kwargs)


    def __repr__(self):
        return '{0} containing\n{1!r}'.format(self.__class__, self.data)

    def __array__(self, *args, **kwargs):
        return self.data.__array__(*args, **kwargs)

class Regions(object):
    REQUIRED_COLUMNS = frozenset(['chromosome', 'start', 'end'])

    _data = None
    def __init__(self, data, *args, **kwargs):

        if isinstance(data, Regions):
            self._data = data.data
        else:
            data = pd.DataFrame(data, *args, **kwargs)
            # Verify that all required columns are in the DF
            for column in self.REQUIRED_COLUMNS:
                if column not in data.columns:
                    raise ValueError('No such column {0!r} in provided DataFrame'.format(column))

            self._data = data

        self._data = self._data.drop_duplicates() # TODO: somehow join the indices of the dropped data

    @property
    def data(self):
        return self._data

    def __repr__(self):
        return '{0} containing \n{1!r}'.format(self.__class__, self.data)


    # --- Initialisation ----------------------------------------------------------------------------------------------
    @classmethod
    def from_bed(cls, bed_file):
        from dgw.data.parsers import read_bed
        return cls(read_bed(bed_file))

    def to_bed(self, bed_file, **track_kwargs):
        from dgw.data.parsers import write_bed
        return write_bed(self, bed_file, **track_kwargs)

    # --- Functions that provide direct access to the DataFrame behind all this ----------------------------------------
    def __getitem__(self, item):
        result = self.data[item]
        if isinstance(result, pd.DataFrame):
            try:
                # Try returining it as the same class
                return self.__class__(result)
            except ValueError:
                # If not valid, return as DataFrame
                return result
        return result

    def has_strand_data(self):
        return 'strand' in self.columns


    def iterrows(self):
        return self.data.iterrows()

    def head(self, *args, **kwargs):
        return self.__class__(self.data.head(*args, **kwargs))

    @property
    def index(self):
        return self.data.index

    def join(self, *args, **kwargs):
        new_data = self.data.join(*args, **kwargs)
        return self.__class__(new_data)

    def append(self, *args, **kwargs):
        new_data = self.data.append(*args, **kwargs)
        return self.__class__(new_data)

    @property
    def ix(self):
        return RegionsIndexer(self.data.ix, self)

    @property
    def columns(self):
        return self.data.columns

    def __getattr__(self, name):
        """
        Emulate the behaviour in `pd.DataFrame` to return one of the columns as an attribute.
        :param name:
        :return:
        """
        # Partially stolen from pandas implementation.
        if name in self.data.columns:
            return self.data[name]

        raise AttributeError("{0!r} has no attribute {1!r}".format(type(self).__name__, name))

    def __len__(self):
        return self.data.__len__()

    # --- Functions special to Regions ---------------------------------------------------------------------------------
    @property
    def lengths(self):
        """
        Returns a `pd.Series` containing the lengths of all the regions contained
        :rtype: `pd.Series`
        """
        series = self['end'] - self['start']
        series.name = 'length'
        return series

    def contained_within(self, other_region):
        """
        Returns all regions that are contained within other region.
        Returns only those regions that are fully contained within the query region.

        :param other_region: The region that regions will be checked to be inside
        :type other_region: `pd.Series`
        :return:
        """
        return self[(self.chromosome == other_region['chromosome'])
                    & (self.start >= other_region['start'])
                    & (self.end <= other_region['end'])]

    def as_printable_list_of_pois(self):

        printable_list = ""
        for ix, row in self.iterrows():
            if printable_list:
                printable_list += ','
            printable_list += ','.join(map(str, range(row['start'], row['end'])))

        return printable_list

    def regions_not_in_dataset(self, dataset):
        """
        Returns all regions that are not in the dataset provided.

        :param dataset:
        :type dataset: AlignmentsData
        :return:
        """
        missing_indices = self.index[~self.index.isin(dataset.items)]
        return Regions(self.data.ix[missing_indices])

    def as_bins_of(self, other_regions, resolution=1, ignore_non_overlaps=False, account_for_strand_information=False):
        """
        Returns the current regions as bins of some other set of regions provided
        :param other_regions: regions to match self to
        :param resolution: resolution at which to do so
        :param ignore_non_overlaps: if set to false, the parser will raise a ValueError if self does not overlap with other regions
        :param account_for_strand_information: if set to true, the parser will account for the antisense regions and
            return bins from the end, rather than front
        :return:
        """
        other_regions = other_regions.clip_to_resolution(resolution)

        bins = {}
        for ix, data in self.iterrows():
            current_start = data['start']
            current_end = data['end']
            current_chromosome = data['chromosome']

            try:
                other = other_regions.ix[ix]
            except KeyError:
                continue

            other_chromosome = other['chromosome']

            if current_chromosome != other_chromosome:
                if ignore_non_overlaps:
                    continue
                else:
                    raise ValueError('Points of interest do not overlap with regions of interest. Failed ix:{0!r}'.format(ix))

            bins[ix] = map_to_bins(range(current_start, current_end), other, resolution=resolution,
                                   ignore_non_overlaps=ignore_non_overlaps,
                                   account_for_strand_information=account_for_strand_information)

        return bins



    def clip_to_resolution(self, resolution):
        """
        Adjusts the region boundaries to fit the resolution by extending the regions boundaries as to length of the
        regions is divisible from resolution.

        Returns a new `Regions` object rather than clipping this one in place

        Please note that the new `Regions` object this function returns will not include any other fields but
        the ones listed in `Regions.REQUIRED_COLUMNS`.

        :param resolution: the resolution to clip
        :type resolution: int
        :return: a new `Regions` object
        :rtype: Regions
        """

        resolution = int(resolution)

        if resolution == 1:
            return self
        elif resolution <= 0:
            raise ValueError('Resolution should be > 0 ({0!r} given)'.format(resolution))

        lengths = self.lengths
        new_regions_data = []

        for ix, row in self.iterrows():
            length = lengths.ix[ix]
            remainder = length % resolution

            if remainder == 0:
                offset_needed = 0
            else:
                offset_needed = resolution - remainder

            add_left = offset_needed / 2
            add_right = offset_needed / 2 + offset_needed % 2

            row['start'] -= add_left

            if row['start'] < 0:
                # Check if we accidentally went sub zero
                add_right += -row['start']
                row['start'] = 0

            row['end'] += add_right

            new_regions_data.append(row)

        df = pd.DataFrame(new_regions_data, index=self.index)
        return Regions(df)

class Genes(Regions):

    def transcription_start_sites(self):
        """
        Returns transcription start site locations for the current set of Genes.
        Pays attention to the strand of the gene and returns the start of the gene for positive strand genes
        and the end of the gene for the negative strand genes.

        :rtype: `pd.Series`
        """
        start_sites = self[self.strand=='+']['start'].append(self[self.strand=='-']['end'])
        start_sites = start_sites.ix[self.index] # reindex the data using original order

        return start_sites

    def regions_around_transcription_start_sites(self, window_width):
        """
        Returns a `Regions` object corresponding to the locations from -window_width to +window_width around the
        transcription start sites for these genes

        :param window_width: the width of the window
        :type window_width: int
        :return: regions around tss
        :rtype: `Regions`
        """

        tss_locations = self.transcription_start_sites()

        # Add window around these
        starts = tss_locations - window_width
        starts[starts < 0] = 0
        ends = tss_locations + window_width

        regions_df = pd.DataFrame({'chromosome' : self.chromosome,
                                   'start' : starts,
                                   'end' : ends,
                                   'strand' : self.strand}, index=self.index)

        return Regions(regions_df)

    def regions_around_splicing_site(self, splicing_site, window_width):
        """
        Return a `Regions` object with windows around first splicing site
        :param splicing_site: number of the splicing site to get (0 based - "0" for first splicing site)
        :param window_width:
        :return:
        """
        exon = self.get_exon_regions(splicing_site)

        exon_genes = self.ix[exon.index]

        ss = exon[exon_genes.strand == '+']['end'].append(exon[exon_genes.strand == '-']['start'])
        ss = ss.ix[self.index]

        starts = ss - window_width
        ends = ss + window_width

        regions_df = pd.DataFrame({'chromosome': exon_genes['chromosome'],
                                  'start': starts, 'end': ends, 'strand' : exon_genes.strand},
                                  index=exon_genes.index)

        return Regions(regions_df)

    def get_exon_regions(self, exon_number, account_for_antisense=True):
        """
        Returns exon regions for particular exon_number provided (0-based)
        :param exon_number: exon number - 0 based. I.e. use 0 to get regions of first exon
        :param account_for_antisense: if set to true, it will automatically account for antisense genes and return correct exon
        :return:
        """
        if 'exonStarts' not in self.columns or 'exonEnds' not in self.columns:
            raise Exception('No exon data available, sorry')

        sub_df = self[['chromosome', 'exonStarts', 'exonEnds', 'strand']]
        new_df = []
        for ix, row in sub_df.iterrows():
            exon_starts = row['exonStarts'].strip(',').split(',')
            exon_ends = row['exonEnds'].strip(',').split(',')
            chromosome = row['chromosome']
            strand = row['strand']

            if account_for_antisense and strand == '-':
                exon_i = len(exon_starts) - 1 - exon_number
            else:
                exon_i = exon_number

            try:
                start = int(exon_starts[exon_i])
                end = int(exon_ends[exon_i])
            except IndexError:
                start = np.nan
                end = np.nan

            new_df.append({'chromosome': chromosome, 'start': start, 'end': end, 'strand': strand})
        new_df = pd.DataFrame(new_df, index=self.index).dropna()
        return Regions(new_df)

    @classmethod
    def from_encode_known_genes(cls, encode_known_genes_filename):
        from dgw.data.parsers import read_encode_known_genes
        return read_encode_known_genes(encode_known_genes_filename)

    @classmethod
    def from_gtf(cls, gtf_filename):
        from dgw.data.parsers import read_gtf
        return read_gtf(gtf_filename)



from collections import defaultdict
from logging import debug
from math import ceil
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.ticker import FuncFormatter, IndexLocator
import numpy as np
import matplotlib
import pandas as pd
from dgw.dtw.utilities import no_nans_len


def dataset_ticks(dataset, scale=1):
    """
    Ticks formater for heatmap. Shows dataset indices as ticks.

    Use it as `ax.yaxis.set_major_formatter(dataset_ticks(dataset, scale))`

    :param dataset:
    :param scale:
    :return:
    """
    if scale is None:
        scale = 1

    def f(tick, pos):
        try:
            return dataset.items[int((tick / scale) + 0.5)]
        except IndexError:
            return ''

    return FuncFormatter(f)

def raw_plot_data_as_heatmap(data_frame, ax=None, highlight_masks=None, highlight_colours=None, *args, **kwargs):
    if ax is None:
        ax = plt.gca()

    # Create a masked array from data_frame
    masked_arr = np.ma.array(data_frame, mask=np.isnan(data_frame))

    # Plot the array using JET colourmap, draw NaNs as white
    cmap = matplotlib.cm.get_cmap('jet')
    cmap.set_bad('#FFFFFF', 1.0)
    HEATMAP_COLORS = ['k', 'w']
    result = ax.imshow(masked_arr, origin='lower', cmap=cmap, aspect="auto", interpolation='nearest', *args, **kwargs)
    if highlight_masks is not None:
        for j, mask in highlight_masks.iteritems():
            highlight_cmap = matplotlib.colors.ListedColormap([highlight_colours[j]])
            highlight_cmap.set_bad('#000000', 0)
            ax.imshow(highlight_masks[j], origin='lower', cmap=highlight_cmap, aspect="auto", interpolation='nearest',
                               *args, **kwargs)

    return result

def plot(alignments, clip_colors=False, titles=None, horizontal_grid=True,
         no_y_axis=False, sort_by=None, subplot_spec=None, share_y_axis=None, scale_y_axis=None, highlighted_points={},
         highlight_colours=None,
         rasterized=True):
    """

    :param alignments: `AlignmentsData` object
    :param clip_colors: will remove some of the highest colours from heatmap for better visualisation if set to true.
                        Equivalent of setting the top 0.1% values in the dataset to the last one below top 0.1%
    :param titles: Titles of the heatmaps. If set to none `alignments.dataset_axis` will be used
    :param horizontal_grid: Whether to plot heatmap on a horizontal grid
    :param no_y_axis: Will not plot the major (items) axis.
    :param sort_by: Sort values by 'length' or supplied index.
    :param subplot_spec: SubplotSpec of the suplot to plot hte heatmaps in. See `matplotlib.gridspec` package.
    :param share_y_axis: if not None, the plot will share the major (items) axis with the specified axis
    :param scale_y_axis: plot will scale the y axis by the specified number (linearly) if set.
    :type scale_y_axis: int
    :param highlighted_points: a (index, array) dictionary of points that should be highlighted in the heatmap
    :param: rasterized: whether to rasterize the plot or not (faster rending for rasterized)
    :return: returns the shared y axis
    """
    import matplotlib.pyplot as plt
    from matplotlib import gridspec

    number_of_datasets = alignments.number_of_datasets

    if titles is None:
        titles = alignments.dataset_axis

    # Should be OK just to just sort the first dataset as all others should have the same len
    sample_data_frame = alignments.dataset_xs(alignments.dataset_axis[0], copy=False).T
    lengths = sample_data_frame.apply(no_nans_len, axis=1)
    max_len = lengths.max()

    # Sorting
    if isinstance(sort_by, pd.Index):
        sorted_index = sort_by
    elif sort_by == 'length':
        debug('Sorting by length')
        lengths.sort()  # Should do in-place
        sorted_index = lengths.index
    elif sort_by is None:
        sorted_index = alignments.items
    else:
        raise ValueError('Unsupported sort_by value provided: {0!r}. Only None or \'length\' supported'.format(sort_by))
    # Apply sorting
    alignments = alignments.ix[sorted_index]
    # Create the instance of axis formatter
    tick_formatter = dataset_ticks(alignments, scale_y_axis)

    # Clipping of colors
    values = np.empty(0)
    for ix in alignments.dataset_axis:
        values = np.concatenate((values, alignments.dataset_xs(ix).unstack().dropna()))

    if clip_colors:
        # Calculate the max_value threshold so heatmap looks good
        histogram, bin_edges = np.histogram(values, bins=1000)

        last_good_bin = None
        TRIM_MAX_VALUE_THRESHOLD = 0.001  # Trim max value threshold to be the value that hides 0.1% of peaks
        allowed_slack = round(TRIM_MAX_VALUE_THRESHOLD * len(values))

        current_slack = 0
        for i, hist in reversed(list(enumerate(histogram))):
            current_slack += hist
            if current_slack >= allowed_slack:
                last_good_bin = i
                break

        max_value = bin_edges[last_good_bin+1]
    else:
        max_value = values.max()

    min_value = 0

    # Subplot creation

    if horizontal_grid:
        grid = (1, number_of_datasets + 1)
        width_ratios = [5] * number_of_datasets
        width_ratios.append(1)
        spacing_kwargs = {'wspace': 0.01, 'width_ratios': width_ratios}  # Almost no space between plots
    else:
        grid = (number_of_datasets, 2)
        spacing_kwargs = {'hspace': 0.15, 'width_ratios': [5, 1]} # Allow a bit of a space for title

    if not subplot_spec:
        gs = gridspec.GridSpec(*grid, **spacing_kwargs)
    else:
        gs = gridspec.GridSpecFromSubplotSpec(*grid, subplot_spec=subplot_spec, **spacing_kwargs)

    # Main drawing loop
    first_axis = None
    extent = None
    if scale_y_axis:
        extent = [0, max_len, 0, alignments.number_of_items * scale_y_axis] # Multiply by 10 as that is what matplotlib's dendrogram returns

    if highlighted_points:
        highlight_masks = defaultdict(lambda: np.zeros((alignments.number_of_items, max_len), dtype=np.bool))
        for i, ix in enumerate(sorted_index):
            try:
                points_of_interest = highlighted_points[ix]
            except KeyError:
                continue

            for j, points in points_of_interest.iteritems():
                highlight_masks[j][i][points] = 1

        for j in highlight_masks.iterkeys():
            highlight_masks[j] = np.ma.masked_where(highlight_masks[j] <= 0, highlight_masks[j])

    else:
        highlight_masks = None

    for i, (ix, title) in enumerate(zip(alignments.dataset_axis, titles)):
        t_gs = gs[:, i] if horizontal_grid else gs[i, 1]
        if i == 0:
            if not share_y_axis:
                first_axis = plt.subplot(t_gs)
                share_y_axis = first_axis
            else:
                first_axis = plt.subplot(t_gs, sharey=share_y_axis)
        else:
            # Remember to share axes
            plt.subplot(t_gs, sharex=first_axis, sharey=share_y_axis)

        plt.gca().get_yaxis().set_major_formatter(tick_formatter)
        #plt.gca().get_yaxis().set_major_locator(IndexLocator(1000, 0))
        # Remove redundant axes
        if horizontal_grid:
            if i > 0 or no_y_axis:
                plt.gca().get_yaxis().set_visible(False)
        else:
            # Leave only last axis
            if i + 1 < number_of_datasets:
                plt.gca().get_xaxis().set_visible(False)

            if no_y_axis:
                plt.gca().get_yaxis().set_visible(False)

        data_to_plot = alignments.dataset_xs(ix, copy=False).T
        # Cut most of the columns that are NaNs out of the plot
        data_to_plot = data_to_plot[data_to_plot.columns[:max_len]]

        result = raw_plot_data_as_heatmap(data_to_plot, vmin=min_value, vmax=max_value, extent=extent,
                                          highlight_masks=highlight_masks, rasterized=rasterized, highlight_colours=highlight_colours)
        debug(plt.gca().get_ylim())
        debug(plt.gca().get_xlim())

        plt.title(title)
        plt.gca().title.set_fontsize(7)
        plt.setp(plt.gca().get_xticklabels(), rotation='vertical', fontsize=7)

    # Colorbar
    colorbar_axis = plt.subplot(gs[:, number_of_datasets] if horizontal_grid else gs[:, 2])
    plt.colorbar(result, cax=colorbar_axis)

    return first_axis


from multiprocessing import cpu_count, Array, Process, Queue
from Queue import Empty, Full
from math import factorial
import itertools
import numpy as np
import ctypes
from logging import debug

from dgw.dtw.distance import dtw_std

__all__ = ['parallel_pdist']

def combinations_count(n_items):
    """
    Returns the number of distinct combinations of n_items there can be.
    In other words, calculates `len(list(itertools.combinations(n_items,2)))` efficiently.
    :param n_items:
    :return:
    """
    return factorial(n_items) / (2 * factorial(n_items - 2))

def _parallel_dtw_worker(data_buffer, operations_generator, data_buffer_shape, result_buffer,
                         scheduling_queue, exception_queue,
                         dtw_args, dtw_kwargs):
    """
    A worker function for parallel_pdist that is executed on a separate process.

    The function takes a shared memory buffer to read the data from.
    A shared memory buffer to save the data into, `result_buffer` is also passed to the function.

    The function takes the regions of the result function to compute from the scheduling queue `scheduling_queue` and performs
    these calculations until the scheduling function becomes empty.

    Any exceptions that occur are pushed to `exception_queue` and the process exits.

    :param data_buffer: shared memory buffer where data is read from
    :param data_buffer_shape: shape of this buffer -- used to convert it to np array
    :param result_buffer: shared memory buffer to store result in
    :param scheduling_queue: multiprocessing-safe queue to read processing schedule from
    :type scheduling_queue: multiprocessing.Queue
    :param exception_queue: queue that any exceptions that occur will be pushed into
    :param dtw_args: args passed into `dtw_std` or `uniform_scaled_distance`
    :param dtw_kwargs: kwargs passed into `dtw_std` or `uniform_scaled_distance`
    :return:
    """
    import sys
    import traceback
    import os

    try:
        pid = os.getpid()
        debug('PROCESS {0}: Spawned'.format(pid))

        data_view = np.frombuffer(data_buffer).reshape(data_buffer_shape)  # Point numpy array to memory
        while True:
            schedule = scheduling_queue.get()
            if schedule is None:
                break  # This indicates there is nothing else will be put in the queue. Kind of a "stop" marker
            else:
                debug('PROCESS {0}: Iteration start'.format(pid))
                start, end = schedule

            # Create the combinations object inside the Process so we can just pass start/end locations in the queue
            # Will need to recreate object every time as we use absolute positions to slice
            combs = operations_generator(xrange(data_buffer_shape[0]))
            data_indices = itertools.islice(combs, start, end)

            # Actual data processing work
            for i, (x, y) in enumerate(data_indices):
                a = data_view[x]
                b = data_view[y]
                result = dtw_std(a, b, *dtw_args, **dtw_kwargs)
                result_buffer[start + i] = result

            debug('PROCESS {0}: Iteration end'.format(pid))


        debug('PROCESS {0}: Work complete'.format(pid))

    except Exception as e:
        exception_queue.put(e)
        traceback.print_exc()
    except:
        e = sys.exc_info()[0]
        exception_queue.put(e)
        traceback.print_exc()

def _pdist_operations_generator_factory(indices):
    return itertools.combinations(indices, 2)

def _parallel_dtw(three_dim_array, _operations_generator_factory, n_operations, n_processes=None, *dtw_args, **dtw_kwargs):
    """
    Runs DTW on parallel
    :param three_dim_array: three-dimensional numpy array of data
    :param _operations_generator_factory: factory function that generates the operations required.
       Should take one argument -- all indices in the data and return a generator of [(i1, j1), (i1,j2), ...]
       where is and js are the operations that need to be computed. See e.g. `_pdist_operations_generator_factory`
    :param n_operations: length of _operations_generator (as generators should not have __len__ method)
    :param n_processes: number of processes to use (defaults to maximum number of CPU cores)
    :param dtw_args: args to pass to dtw
    :param dtw_kwargs: kwargs to pass to dtw
    :return:
    """

    three_dim_array = np.asarray(three_dim_array)
    if n_processes is None:
        n_processes = cpu_count()
    else:
        n_processes = int(n_processes)
        if n_processes <= 0:
            raise ValueError('N_processes should be > 0')
        elif n_processes > cpu_count():
            raise ValueError('The specified number of CPUs to use, {0} is greater than the number of available CPUs, {1}'
            .format(n_processes, cpu_count()))

    debug('Using {0} processes for parallel computation'.format(n_processes))

    # Create a shared memory array to store result
    # Do not lock it as the worker should make sure processes do not overlap the data
    result_buffer = Array(ctypes.c_double, n_operations, lock=False)

    # Create a shared memory buffer for the data array
    shape = three_dim_array.shape
    data_buffer = Array(ctypes.c_double, np.product(shape), lock=False)  # Allocate memory
    data_view = np.frombuffer(data_buffer).reshape(shape)  # Point numpy array to memory
    data_view[:] = three_dim_array  # Copy the contents into the new memory location

    number_of_slices = n_processes * 4

    buffer_size, remainder = divmod(n_operations, number_of_slices)

    scheduling_queue = Queue()

    # Split the data in slices
    for i in xrange(number_of_slices):
        start = i * buffer_size
        end = start + buffer_size
        scheduling_queue.put((start, end))
    if remainder:
        last_start = number_of_slices * buffer_size
        last_end = n_operations
        scheduling_queue.put((last_start, last_end))

    for i in xrange(n_processes):
        scheduling_queue.put(None)  # Add stop items to the queue so we know when its empty for sure

    # Create queue for exceptions
    exception_queue = Queue()

    processes = []
    for i in xrange(n_processes):
        p = Process(target=_parallel_dtw_worker,
                    args=(data_buffer, _operations_generator_factory, shape, result_buffer, scheduling_queue,
                          exception_queue, dtw_args, dtw_kwargs))
        processes.append(p)

    # Start all processes
    for p in processes:
        p.start()

    # Join all processes
    for p in processes:
        p.join()

    try:
        exception = exception_queue.get(block=False)
        raise exception
    except Empty:
        pass

    # Convert the result to numpy array in the end
    return np.frombuffer(result_buffer)

def parallel_pdist(three_dim_array, n_processes=None, *dtw_args, **dtw_kwargs):
    """
    Calculates pairwise DTW distance for all the rows in three_dim_array provided.
    This module is similar to scipy.spatial.distance.pdist, but uses all CPU cores available, rather than one.
    :param three_dim_array: numpy data array [observations x max(sequence_lengths) x ndim ]
    :param n_processes: number of processes to spawn usage to the number specified.
                        Will default to the number of (virtual) CPUs available if not set
    :param dtw_args: `args` to be passed into `dtw_std`
    :param dtw_kwargs: `kwargs` to be passed into `dtw_std`
    :return: condensed distance matrix (just as `scipy.spatial.distance.pdist`)
    """
    three_dim_array = np.asarray(three_dim_array)
    n_items = three_dim_array.shape[0]
    number_of_combinations = combinations_count(n_items)

    return _parallel_dtw(three_dim_array, _pdist_operations_generator_factory, number_of_combinations,
                         n_processes=n_processes, *dtw_args, **dtw_kwargs)


def _path_calculation_worker(data_buffer, shape, prototypes_buffer, prototypes_shape,
                             scheduling_queue, results_queue, exception_queue,
                             dtw_args, dtw_kwargs):
    import sys
    import traceback
    import os

    try:
        pid = os.getpid()
        debug('PROCESS {0}: Spawned'.format(pid))
        data_view = np.frombuffer(data_buffer).reshape(shape)  # Point numpy array to memory

        prototypes_view = np.frombuffer(prototypes_buffer).reshape(prototypes_shape)
        answers = []
        while True:
            work_id = scheduling_queue.get()
            if work_id is None:
                break
            data_i, base_i = work_id
            x = data_view[data_i]
            base = prototypes_view[base_i]

            _, _, path = dtw_std(x, base, dist_only=False, *dtw_args, **dtw_kwargs)

            results_queue.put((work_id, path))

        debug('PROCESS {0}: finished computation'.format(pid))
        debug('PROCESS {0}: finished'.format(pid))

    except Exception as e:
        exception_queue.put(e)
        traceback.print_exc()
    except:
        e = sys.exc_info()[0]
        exception_queue.put(e)
        traceback.print_exc()

def parallel_dtw_paths(full_data, nodes, n_processes=None, *dtw_args, **dtw_kwargs):
    def _read_results_of_queue_till_empty(queue, ans_dict):
        debug('Queue full, reading results till can proceed further')
        number_of_answers = 0
        while number_of_answers < 100:
            try:
                (i, j), ps = queue.get(block=False)
            except Empty:
                debug('Answers queue empty, continuing')
                break
            ans_dict[node_id_lookup[j]][data_index[i]] = ps
            number_of_answers += 1

        debug('Got {0} answers back'.format(number_of_answers))
        return number_of_answers


    if n_processes is None:
        n_processes = cpu_count()
    else:
        n_processes = int(n_processes)
        if n_processes <= 0:
            raise ValueError('N_processes should be > 0')
        elif n_processes > cpu_count():
            raise ValueError('The specified number of CPUs to use, {0} is greater than the number of available CPUs, {1}'
            .format(n_processes, cpu_count()))

    debug('Using {0} processes for parallel computation'.format(n_processes))

    max_prototype_len = 0
    for node in nodes:
        prototype_len = len(node.prototype)
        if prototype_len > max_prototype_len:
            max_prototype_len = prototype_len
    ndims = full_data.values.shape[2]
    nan = [np.nan] * ndims
    prototypes = np.empty((len(nodes), max_prototype_len, ndims))
    for i, node in enumerate(nodes):
        prototype = node.prototype
        prototype_len = len(prototype)
        if prototype_len < max_prototype_len:
            padding = np.asarray([nan] * (max_prototype_len - prototype_len))
            padded_prototype = np.concatenate((prototype, padding))
            prototypes[i] = padded_prototype
        else:
            prototypes[i] = prototype

    # Create shared memory buffer fro prototypes array
    prototypes_shape = prototypes.shape
    prototypes_buffer = Array(ctypes.c_double, np.product(prototypes_shape), lock=False)
    prototypes_view = np.frombuffer(prototypes_buffer).reshape(prototypes_shape)
    prototypes_view[:] = prototypes

    # Generate lookup for indices
    data_index = full_data.items
    ix_lookup = {}
    for i, ix in enumerate(data_index):
        ix_lookup[ix] = i

    node_id_lookup = {}
    for j, node in enumerate(nodes):
        node_id_lookup[j] = node.id

    full_data = np.asarray(full_data)

    # Create a shared memory buffer for the data array
    shape = full_data.shape
    data_buffer = Array(ctypes.c_double, np.product(shape), lock=False)  # Allocate memory
    data_view = np.frombuffer(data_buffer).reshape(shape)  # Point numpy array to memory
    data_view[:] = full_data  # Copy the contents into the new memory location


    answers_queue = Queue()
    exception_queue = Queue()

    scheduling_queue = Queue()
    processes = []
    for i in xrange(n_processes):
        p = Process(target=_path_calculation_worker,
                    args=(data_buffer, shape, prototypes_buffer, prototypes_shape, scheduling_queue, answers_queue,
                          exception_queue, dtw_args, dtw_kwargs))
        processes.append(p)

    # Buffer to store results
    paths = {}
    for node in nodes:
        paths[node.id] = {}

    # Start all processes
    for p in processes:
        debug('Starting process')
        p.start()

    # Keep pushing things for the processes to work on
    n_items = 0

    number_of_answers = 0

    for j, node in enumerate(nodes):
        indices = node.index

        for ix in indices:
            while True:
                try:
                    scheduling_queue.put((ix_lookup[ix], j), block=False)
                    n_items += 1
                    break
                except Full:
                    # Whoops, our scheduling queue is full
                    debug('Full at {0} sent'.format(n_items))
                    number_of_answers += _read_results_of_queue_till_empty(answers_queue, paths)
                    continue

    # Stop all processes
    for i in xrange(n_processes):
        while True:
            try:
                scheduling_queue.put(None, block=False)
                break
            except Full:
                debug('Full while sending the Nones')
                number_of_answers += _read_results_of_queue_till_empty(answers_queue, paths)
                continue

    debug('Finished scheduling tasks, reading off results')
    while number_of_answers < n_items:
        (i, j), ps = answers_queue.get()
        paths[node_id_lookup[j]][data_index[i]] = ps
        number_of_answers += 1

    # Join all processes
    for p in processes:
        debug('Joining process')
        p.join()

    debug('Checking exceptions')
    try:
        exception = exception_queue.get(block=False)
        raise exception
    except Empty:
        pass

    return paths


import unittest

import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from dgw.dtw import parametrised_dtw_wrapper, uniform_scaling_to_length, reverse_sequence
from dgw.dtw.scaling import uniform_shrinking_to_length
from dgw.dtw.transformations import *


class TestScaling(unittest.TestCase):

    def test_uniform_extending_to_length_one_dim(self):
        a = np.array([1, 2, 3])

        # Easy cases - new dimension is a multiple of previous
        assert_array_equal(np.array([1, 1, 2, 2, 3, 3]), uniform_scaling_to_length(a, 6))

        assert_array_equal(np.array([1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3]), uniform_scaling_to_length(a, 12))

        # Harder cases - new dimension is not a multiple of previous
        assert_array_equal(np.array([1, 1, 2, 3]), uniform_scaling_to_length(a, 4))
        assert_array_equal(np.array([1, 1, 2, 2, 3]), uniform_scaling_to_length(a, 5))

    def test_uniform_extending_to_length_two_dim(self):

        a = np.array([[1, 4], [2, 5], [3, 6]])

        # Easy cases -- new dimension is a multiple of previous
        assert_array_equal(np.array([[1, 4], [1, 4], [2, 5], [2, 5], [3, 6], [3, 6]]),
                           uniform_scaling_to_length(a, 6))

        assert_array_equal(np.array([[1, 4], [1, 4], [1, 4], [1, 4],
                                     [2, 5], [2, 5], [2, 5], [2, 5],
                                     [3, 6], [3, 6], [3, 6], [3, 6]]),
                           uniform_scaling_to_length(a, 12))

        # Harder ases - new dimension is not a multiple of previous
        assert_array_equal(np.array([[1, 4], [1, 4], [2, 5], [3, 6]]), uniform_scaling_to_length(a, 4))
        assert_array_equal(np.array([[1, 4], [1, 4], [2, 5], [2, 5], [3, 6]]), uniform_scaling_to_length(a, 5))
        
    def test_uniform_extending_with_nans(self):
        a = np.array([1, 2, 3, np.nan])
        assert_array_equal(np.array([1, 1, 2, 2, 3, 3]), uniform_scaling_to_length(a, 6))

        b = np.array([[1, 4], [2, 5], [3, 6], [np.nan, np.nan]])
        assert_array_equal(np.array([[1, 4], [1, 4], [2, 5], [2, 5], [3, 6], [3, 6]]),
                           uniform_scaling_to_length(b, 6))

    def test_uniform_extending_edge_case_one_element(self):
        a = np.array([1])
        assert_array_equal(np.array([1, 1, 1, 1]), uniform_scaling_to_length(a, 4))

        a2 = np.array([1, np.nan])
        assert_array_equal(np.array([1, 1, 1, 1]), uniform_scaling_to_length(a2, 4))

        b = np.array([[1, 17]])
        assert_array_equal(np.array([[1, 17], [1, 17], [1, 17], [1, 17]]), uniform_scaling_to_length(b, 4))

        b2 = np.array([[1, 17], [np.nan, np.nan]])
        assert_array_equal(np.array([[1, 17], [1, 17], [1, 17], [1, 17]]), uniform_scaling_to_length(b2, 4))

    def test_uniform_extending_edge_case_no_elements(self):
        a = np.array([])
        self.assertRaises(ValueError, uniform_scaling_to_length, a, 4)

        b = np.array([np.nan])
        self.assertRaises(ValueError, uniform_scaling_to_length, b, 4)

    def test_uniform_extending_fails_when_shrinking_asked_for(self):
        a = np.array([1, 2, 3])
        self.assertRaises(ValueError, uniform_scaling_to_length, a, 2)

    def test_uniform_extending_does_nothing_when_length_is_the_same(self):
        a = np.array([1, 2, 3])
        assert_array_equal(a, uniform_scaling_to_length(a, len(a)))

        b = np.array([[1, 4], [2, 5], [3, 6]])
        assert_array_equal(b, uniform_scaling_to_length(b, len(b)))

    def test_uniform_shrinking_to_length_one_dim(self):
        a = np.array([1, 2, 3, 4, 5, 6])

        # Easy cases each item is an average of the blocks
        assert_array_equal(np.array([1.5, 3.5, 5.5]), uniform_shrinking_to_length(a, 3))
        assert_array_equal(np.array([2, 5]), uniform_shrinking_to_length(a, 2))
        assert_array_equal(np.array([3.5]), uniform_shrinking_to_length(a, 1))

        # Harder cases
        b = np.array([1, 2, 3, 4, 5, 6, 7])

        # When scaling from size 7 to size 3
        # each scaled element will consist of 7/3 = 2.333 elements in longer sequence:
        # Thus short[0] is equivalent to average between 0 and 2.333 elements in the long sequence
        #      short[1] is average of elements 2.333 - 4.666 in the long seq
        #      short[2] is average of elements 4.666 - 7 in the long seq

        # Average of fractional elements can be computed by assuming the element "contributes" to average
        # only fractionally e.g. if the interval is [1.2, 3.6) then the avg should be computed as:
        # (1-0.2) * 2 + 1 * 3 + 0.6 * 4 / (0.8 + 1 + 0.6)

        ans = np.array([(1 + 2 + (1 / 3.0) * 3) / (1 + 1 + (1 / 3.0)),  # [0, 2.333)
                        ((1 - (1 / 3.0)) * 3 + 4 + (2 / 3.0) * 5) / (1 - (1 / 3.0) + 1 + (2 / 3.0)),  # [2.333; 4.666]
                        ((1 - (2 / 3.0)) * 5 + 6 + 7) / (1 - (2 / 3.0) + 1 + 1)])  # [4.666; 7)

        result = uniform_shrinking_to_length(b, 3)

        assert_array_almost_equal(ans, result)


    def test_uniform_shrinking_to_length_two_dim(self):
        a = np.array([[1, 7], [2, 8], [3, 9], [4, 10], [5, 11], [6, 12]])

        # Easy cases each item is an average of the blocks
        assert_array_equal(np.array([[1.5, 7.5], [3.5, 9.5], [5.5, 11.5]]), uniform_shrinking_to_length(a, 3))
        assert_array_equal(np.array([[2, 8], [5, 11]]), uniform_shrinking_to_length(a, 2))
        assert_array_equal(np.array([[3.5, 9.5]]), uniform_shrinking_to_length(a, 1))

        # Harder cases
        b = np.array([[1, 8], [2, 9], [3, 10], [4, 11], [5, 12], [6, 13], [7, 14]])

        # When scaling from size 7 to size 3
        # each scaled element will consist of 7/3 = 2.333 elements in longer sequence:
        # Thus short[0] is equivalent to average between 0 and 2.333 elements in the long sequence
        #      short[1] is average of elements 2.333 - 4.666 in the long seq
        #      short[2] is average of elements 4.666 - 7 in the long seq

        # Average of fractional elements can be computed by assuming the element "contributes" to average
        # only fractionally e.g. if the interval is [1.2, 3.6) then the avg should be computed as:
        # (1-0.2) * 2 + 1 * 3 + 0.6 * 4 / (0.8 + 1 + 0.6)

        one_third = 1 / 3.0
        two_thirds = 2 / 3.0
        ans = np.array([[(1 + 2 + one_third * 3) / (1 + 1 + one_third),
                             (8 + 9 + one_third * 10.0) / (1 + 1 + one_third)], # [0, 2.333)
                        [((1 - one_third) * 3 + 4 + two_thirds * 5) / (1 - one_third + 1 + two_thirds),
                             ((1 - one_third) * 10 + 11 + two_thirds * 12) / (1 - one_third + 1 + two_thirds)],  # [2.333; 4.666]
                        [((1 - two_thirds) * 5 + 6 + 7) / (1 - two_thirds + 1 + 1),
                             ((1 - two_thirds) * 12 + 13 + 14) / (1 - two_thirds + 1 + 1)]])  # [4.666; 7)

        result = uniform_shrinking_to_length(b, 3)
        assert_array_almost_equal(ans, result)

    def test_uniform_shrinking_to_length_raises_exception_when_extension_needed(self):

        self.assertRaises(ValueError, uniform_shrinking_to_length, np.array([1, 2, 3]), 6)
        self.assertRaises(ValueError, uniform_shrinking_to_length, np.array([1, 2, 3, np.nan, np.nan]), 4)
        self.assertRaises(ValueError, uniform_shrinking_to_length, np.array([np.nan, np.nan, np.nan, np.nan]), 3)

    def test_uniform_shrinking_does_nothing_when_lengths_equal(self):
        a = np.array([1, 2, 3])
        assert_array_equal(a, uniform_shrinking_to_length(a, 3))

        b = np.array([1, 2, 3, np.nan, np.nan])
        assert_array_equal(a, uniform_shrinking_to_length(b, 3))

    def test_uniform_shrinking_raises_exception_on_invalid_length(self):
        self.assertRaises(ValueError, uniform_shrinking_to_length, np.array([1, 2, 3]), 0)


    def test_uniform_shrinking_doesnt_crash_when_doing_55_45(self):
        # This test is here to make sure it does not crash when the final boundary is affected by rounding errors
        a = np.ones(55)
        assert_array_equal(np.ones(45), uniform_shrinking_to_length(a, 45))


class TestPathAveraging(unittest.TestCase):

    def test_one_dim(self):
        a = np.array([1, 2, 3])
        b = np.array([5, 8, 9, 11])

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        correct_ans = np.array([3, 5, 6, 7])
        average_path = dtw_path_averaging(a, b, path=path)
        assert_array_equal(correct_ans, average_path)

        path2 = (np.array([0, 1, 1, 2]),
                 np.array([0, 1, 2, 3]))

        correct_ans2 = np.array([3, 5, (2 + 9) / 2.0, 7])
        average_path2 = dtw_path_averaging(a, b, path=path2)
        assert_array_equal(correct_ans2, average_path2)

    def test_weighted_one_dim(self):
        a = np.array([1, 2, 3])
        b = np.array([5, 8, 9, 11])

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        correct_ans = np.array([(2 * 1 + 4 * 5) / 6.0,
                                (2 * 2 + 4 * 8) / 6.0,
                                (2 * 3 + 4 * 9) / 6.0,
                                (2 * 3 + 4 * 11) / 6.0])

        average_path = dtw_path_averaging(a, b, 2, 4, path=path)
        assert_array_equal(correct_ans, average_path)

        path2 = (np.array([0, 1, 1, 2]),
                 np.array([0, 1, 2, 3]))

        correct_ans2 = np.array([(2 * 1 + 4 * 5) / 6.0,
                                (2 * 2 + 4 * 8) / 6.0,
                                (2 * 2 + 4 * 9) / 6.0,
                                (2 * 3 + 4 * 11) / 6.0])
        
        average_path2 = dtw_path_averaging(a, b, 2, 4, path=path2)
        assert_array_equal(correct_ans2, average_path2)

    def test_multi_dim(self):

        a = np.array([[1, 21], [2, 22], [3, 23]])
        b = np.array([[5, 35], [8, 38], [9, 39], [11, 41]])

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        correct_ans = np.array([[3, (35 + 21) / 2.0],
                                [5, (38 + 22) / 2.0],
                                [6, ((39 + 23) / 2.0)],
                                [7, (41 + 23) / 2.0]])

        average_path = dtw_path_averaging(a, b, path=path)
        assert_array_equal(correct_ans, average_path)

        path2 = (np.array([0, 1, 1, 2]),
                 np.array([0, 1, 2, 3]))

        correct_ans2 = np.array([[3, (35 + 21) / 2.0],
                                [5, (38 + 22) / 2.0],
                                [(2 + 9) / 2.0, ((39 + 22) / 2.0)],
                                [7, (41 + 23) / 2.0]])
        average_path2 = dtw_path_averaging(a, b, path=path2)
        assert_array_equal(correct_ans2, average_path2)

class TestProjection(unittest.TestCase):

    def setUp(self):
        np.random.seed(42)

    def test_dtw_std_is_the_same_regardless_of_whether_path_is_computed_or_provided(self):
        a = np.random.randn(20)
        b = np.random.randn(30)
        dtw_function = dtw_std

        dist, cost, path = dtw_function(a, b, dist_only=False)
        path_answer = dtw_projection(a, b, path=path)
        dtw_function_answer = dtw_projection(a, b, dtw_function=dtw_function)
        assert_array_equal(path_answer, dtw_function_answer)


    def test_dtw_sb_is_the_same_regardless_of_whether_path_is_computed_or_provided(self):
        np.random.seed(42)
        a = np.random.randn(20)
        b = np.random.randn(30)

        dtw_function = parametrised_dtw_wrapper(constraint='slanted_band', k=4)

        dist, cost, path = dtw_function(a, b, dist_only=False)
        path_answer = dtw_projection(a, b, path=path)
        dtw_function_answer = dtw_projection(a, b, dtw_function=dtw_function)
        assert_array_equal(path_answer, dtw_function_answer)

    def test_projection_is_the_same_regardless_of_whether_input_is_reversed(self):
        a = np.random.randn(10, 5)
        b = np.random.randn(20, 5)

        dtw_function = parametrised_dtw_wrapper(try_reverse=True)
        ans_norm = dtw_projection(a, b, dtw_function)
        ans_reverse = dtw_projection(reverse_sequence(a), b, dtw_function)

        assert_array_almost_equal(ans_norm, ans_reverse)

    def test_projection_is_reversed_when_base_is_reversed(self):
        a = np.arange(10, 20, 1)
        b = np.arange(20, 50, 1)

        dtw_function = parametrised_dtw_wrapper(try_reverse=True)
        ans_norm = dtw_projection(a, b, dtw_function)
        ans_reverse = dtw_projection(a, reverse_sequence(b), dtw_function)

        assert_array_equal(reverse_sequence(ans_norm), ans_reverse)

class TestSdtwAveraging(unittest.TestCase):

    def test_one_dim_weights_set_to_one(self):
        a = np.array([1, 2, 3])
        b = np.array([5, 8, 9, 11])

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        # Should be the same as scaled regular path averaging when weights = 1
        correct_ans = uniform_shrinking_to_length(np.array([3, 5, 6, 7]), 4)
        average_path = sdtw_averaging(a, b, 1, 1, path=path)

        assert_array_equal(correct_ans, average_path)

        path2 = (np.array([0, 1, 1, 2]),
                 np.array([0, 1, 2, 3]))

        correct_ans2 = uniform_shrinking_to_length(np.array([3, 5, (2 + 9) / 2.0, 7]), 4)
        average_path2 = sdtw_averaging(a, b, 1, 1, path=path2)
        assert_array_equal(correct_ans2, average_path2)

    def test_multi_dim_weights_set_to_one(self):

        a = np.array([[1, 21], [2, 22], [3, 23]])
        b = np.array([[5, 35], [8, 38], [9, 39], [11, 41]])

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        correct_ans = np.array([[3, (35 + 21) / 2.0],
                                [5, (38 + 22) / 2.0],
                                [6, ((39 + 23) / 2.0)],
                                [7, (41 + 23) / 2.0]])

        correct_ans = uniform_shrinking_to_length(correct_ans, 4)
        average_path = sdtw_averaging(a, b, 1, 1, path=path)
        assert_array_equal(correct_ans, average_path)

        path2 = (np.array([0, 1, 1, 2]),
                 np.array([0, 1, 2, 3]))

        correct_ans2 = np.array([[3, (35 + 21) / 2.0],
                                 [5, (38 + 22) / 2.0],
                                 [(2 + 9) / 2.0, ((39 + 22) / 2.0)],
                                 [7, (41 + 23) / 2.0]])
        correct_ans2 = uniform_shrinking_to_length(correct_ans2, 4)
        average_path2 = sdtw_averaging(a, b, 1, 1, path=path2)
        assert_array_equal(correct_ans2, average_path2)

    def test_one_dim_weights_more_than_one(self):
        a = np.array([1, 2, 3])
        b = np.array([5, 8, 9, 11])

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        # Calculate weighted averages for elements between sequences
        a0 = float(1 * 3 + 5 * 7) / (3 + 7)
        a1 = float(2 * 3 + 8 * 7) / (3 + 7)
        a2 = float(3 * 3 + 9 * 7) / (3 + 7)
        a3 = float(3 * 3 + 11 * 7) / (3 + 7)

        correct_ans = np.array([a0, a0, a0, a0, a0,  # Diagonal
                                a1, a1, a1, a1, a1,  # Diagonal step
                                a2, a2, a2, a2, a2,  # Diagonal step
                                a3, a3, a3])       # B moved only

        correct_ans = uniform_shrinking_to_length(correct_ans, 4)
        average_path = sdtw_averaging(a, b, 3, 7, path=path)
        assert_array_equal(correct_ans, average_path)


    def test_multi_dim_weights_set_to_more_than_one(self):

        a = np.array([[1, 21], [2, 22], [3, 23]], dtype=float)
        b = np.array([[5, 35], [8, 38], [9, 39], [11, 41]], dtype=float)

        path = (np.array([0, 1, 2, 2]),
                np.array([0, 1, 2, 3]))

        # Calculate weighted averages for elements between sequences
        a0 = float(1 * 3 + 5 * 7) / (3 + 7)
        a1 = float(2 * 3 + 8 * 7) / (3 + 7)
        a2 = float(3 * 3 + 9 * 7) / (3 + 7)
        a3 = float(3 * 3 + 11 * 7) / (3 + 7)

        b0 = float(21 * 3 + 35 * 7) / (3 + 7)
        b1 = float(22 * 3 + 38 * 7) / (3 + 7)
        b2 = float(23 * 3 + 39 * 7) / (3 + 7)
        b3 = float(23 * 3 + 41 * 7) / (3 + 7)

        correct_ans = np.array([[a0, b0], [a0, b0], [a0, b0], [a0, b0], [a0, b0],
                                [a1, b1], [a1, b1], [a1, b1], [a1, b1], [a1, b1],
                                [a2, b2], [a2, b2], [a2, b2], [a2, b2], [a2, b2],
                                [a3, b3], [a3, b3], [a3, b3]], dtype=float)

        average_path = sdtw_averaging(a, b, 3, 7, path=path, shrink=False)
        assert_array_equal(correct_ans, average_path, 'Arrays not equal before shrinking')

        shrinked_ans = uniform_shrinking_to_length(correct_ans, 4)
        average_shrinked_path = sdtw_averaging(a, b, 3, 7, path=path, shrink=True)
        assert_array_equal(shrinked_ans, average_shrinked_path, 'Arrays not equal after shrinking')




import numpy as np

def map_to_bins(poi, other_region, resolution=50, ignore_non_overlaps=False, account_for_strand_information=False):
    """
    Maps the points of interest to the bins in the other region provided.
    Note that this does not validate the chromosome, this needs to be done by the caller

    :param poi:
    :param other_region:
    :param resolution:
    :param ignore_non_overlaps:
    :param account_for_strand_information:
    :return:
    """

    other_start = other_region['start']
    other_end = other_region['end']

    poi = np.asarray(poi, dtype=int)

    if np.min(poi) < other_start or np.max(poi) >= other_end:
        if ignore_non_overlaps:
            return []
        else:
            raise ValueError('Points of interest do not overlap with regions of interest.')
    resolution = int(resolution)
    n_bins = (other_end - other_start) / resolution
    poi_offsetted = poi - other_start
    poi_bins = poi_offsetted / resolution  # This should be integer division

    # Remove duplicates
    poi_bins = np.unique(poi_bins)

    if not account_for_strand_information or other_region['strand'] == '+':
        return poi_bins
    else:
        return sorted((n_bins - 1) - poi_bins)

def from_simple(poi_file, main_regions, resolution=50, ignore_non_overlaps=False, account_for_strand_information=False):
    """
    Parses points of interest from the specified file in the simple form:
    ```index:poi_1,poi_2,poi_3,...,poi_n```, e.g.
    ```MACS_peak_1:23,45,100```

    :param poi_file: file with pois in simple format
    :param main_regions: the main regions the pois will be parsed from
    :return:
    """

    if isinstance(poi_file, basestring):
        poi_file = open(poi_file)
        need_closing = True
    else:
        need_closing = False

    try:
        poi_dict = {}

        for line in poi_file:
            try:
                ix, pois = line.split(':')
            except ValueError:
                raise ValueError('Not a valid poi file in simple format')

            pois = map(int, pois.replace(' ', '').split(','))

            try:
                main_region = main_regions.ix[ix]
            except KeyError:
                # Some of the regions in POI might not exist in main regions, e.g. due to --random-sample constraint ...
                # that's okay
                continue

            bins = map_to_bins(pois, main_region, resolution=resolution, ignore_non_overlaps=ignore_non_overlaps,
                               account_for_strand_information=account_for_strand_information)

            if len(bins) > 0:
                poi_dict[ix] = bins
        return poi_dict
    finally:
        if need_closing:
            poi_file.close()






## This code is written by Davide Albanese, <albanese@fbk.eu> (C) 2011 mlpy Developers
##  and later modified by Saulius Lukauskas, <luksaulius@gmail.com>

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
cimport numpy as np

from libc.stdlib cimport *
from cdtw cimport *

np.import_array()

cdef retrace_path(int n, int m, np.ndarray[np.float_t, ndim=2] cost_arr):
    '''
       Retraces the warping path back from cost_arr.

    :param n: length of the first sequence
    :param m: length of the second sequence
    :param cost_arr: the computed cost array
    :return: (px_arr, py_arr) - warping path for both sequences.
    '''
    cdef Path p
    cdef np.ndarray[np.int_t, ndim=1] px_arr
    cdef np.ndarray[np.int_t, ndim=1] py_arr

    path(<double *> cost_arr.data,
        n, m,
        -1, -1, &p)
    px_arr = np.empty(p.k, dtype=np.int)
    py_arr = np.empty(p.k, dtype=np.int)
    for i in range(p.k):
        px_arr[i] = p.px[i]
        py_arr[i] = p.py[i]
    free (p.px)
    free (p.py)

    return px_arr, py_arr

def dtw_std(x, y, dist_only=True, metric='euclidean', constraint=None, k=None, warping_penalty=0):
    """Standard DTW as described in [Muller07]_,
    using the Euclidean distance (absolute value 
    of the difference) or squared Euclidean distance
    (as in [Keogh01]_) as local cost measure.

    :Parameters:
       x : 1d array_like object (N)
          first sequence
       y : 1d array_like object (M)
          second sequence
       dist_only : bool
          compute only the distance
       metric : 'euclidean', 'sqeuclidean' or 'cosine'
          distance metric to use
       constraint: string
          one of the following:
             None or ('None') : unconstrained DTW.
             'sakoe_chiba': DTW constrained by Sakoe & Chiba band of width 2k + 1 (requires value of k set), see [Sakoe78]
             'slanted_band': Generalisation of Sakoe & Chiba constraint that supports sequences of different lengths
             'itakura'    : DTW constrained by Itakura Parallelogram, see
       k : int
          parameter required by sakoe_chiba and slanted_band constraints.
       warping_penalty: double
          warping penalty to impose on non-diagonal path changes (default: 0)
       :Returns:
       dist : float
          unnormalized minimum-distance warp path 
          between sequences
       cost : 2d numpy array (N,M) [if dist_only=False]
          accumulated cost matrix
       path : tuple of two 1d numpy array (path_x, path_y) [if dist_only=False]
          warp path
    
    .. [Muller07] M Muller. Information Retrieval for Music and Motion. Springer, 2007.
    .. [Keogh01] E J Keogh, M J Pazzani. Derivative Dynamic Time Warping. In First SIAM International Conference on Data Mining, 2001.
    .. [Sakoe78] H Sakoe, & S Chiba S. Dynamic programming algorithm optimization for spoken word recognition. Acoustics, 1978
    .. [Itakura75] F Itakura. Minimum prediction residual principle applied to speech recognition. Acoustics, Speech and Signal Processing, IEEE Transactions on, 23(1), 67–72, 1975. doi:10.1109/TASSP.1975.1162641.
    """

    x = np.ascontiguousarray(x, dtype=np.float)
    y = np.ascontiguousarray(y, dtype=np.float)

    if x.ndim == 1 and x.ndim == 1: # Turn one-dimensional array into two-dimensional one
        x = np.reshape(x, (-1, 1))
        y = np.reshape(y, (-1, 1))

    if x.shape[1] != y.shape[1]:
        raise ValueError('Both sequences must have the same number of dimensions in each element')


    cdef np.ndarray[np.float_t, ndim=2] x_arr
    cdef np.ndarray[np.float_t, ndim=2] y_arr
    cdef np.ndarray[np.float_t, ndim=2] cost_arr

    cdef double dist
    cdef int i

    cdef np.ndarray[np.int_t, ndim=1] px_arr
    cdef np.ndarray[np.int_t, ndim=1] py_arr

    x_arr = np.ascontiguousarray(x, dtype=np.float)
    y_arr = np.ascontiguousarray(y, dtype=np.float)

    cdef int n = x_arr.shape[0]
    cdef int m = y_arr.shape[0]
    cdef int n_dimensions = x_arr.shape[1]

    cost_arr = np.empty((n,m), dtype=np.float)

    cdef int distance

    if metric == 'sqeuclidean':
        distance = MLPY_DTW_DISTANCE_SQEUCLIDEAN
    elif metric == 'euclidean':
        distance = MLPY_DTW_DISTANCE_EUCLIDEAN
    elif metric == 'cosine':
        distance = MLPY_DTW_DISTANCE_COSINE
    else:
        raise ValueError('Unsupported distance metric provided: {0!r}.'.format(metric))

    if constraint is None or constraint == 'None' or constraint == 'none':
        fill_cost_matrix_unconstrained(
            <double *> x_arr.data, <double *> y_arr.data,
            <int> n, <int> m, <int> n_dimensions,
            distance, warping_penalty, <double *> cost_arr.data)
    elif constraint == 'sakoe_chiba':
        if k is None:
            raise ValueError('Please specify value of k for Sakoe & Chiba constraint')

        k = int(k)
        if k < 0:
            raise ValueError('Value of k must be greater or equal than 0')

        fill_cost_matrix_with_sakoe_chiba_constraint(
            <double *> x_arr.data, <double *> y_arr.data,
            <int> n, <int> m, <int> n_dimensions,
            distance, warping_penalty, <double *> cost_arr.data,
            <int> k
        )
    elif constraint == 'slanted_band':
        if k is None:
            raise ValueError('Please specify value of k for Slanted Band constraint')

        k = int(k)
        if k < 0:
            raise ValueError('Value of k must be greater or equal than 0')

        transpose_cost = False
        if n < m:
            # swap the arrays as the cost matrix filling function expects len(x) > len(y)
            n, m = m, n
            x_arr, y_arr = y_arr, x_arr
            cost_arr = np.asfortranarray(cost_arr)
            transpose_cost = True

        fill_cost_matrix_with_slanted_band_constraint(
            <double *> x_arr.data, <double *> y_arr.data,
            <int> n, <int> m, <int> n_dimensions,
            distance, warping_penalty, <double *> cost_arr.data,
            <int> k
        )

        if transpose_cost:
            # swap everything back again
            cost_arr = np.ascontiguousarray(cost_arr)
            x_arr, y_arr = y_arr, x_arr
            n, m = m, n

    elif constraint == 'itakura':
        fill_cost_matrix_with_itakura_constraint(
            <double *> x_arr.data, <double *> y_arr.data,
            <int> n, <int> m, <int> n_dimensions,
            distance, warping_penalty, <double *> cost_arr.data)
    dist = cost_arr[n-1, m-1]

    if dist_only:
        return dist
    else:
        px_arr, py_arr = retrace_path(x_arr.shape[0], y_arr.shape[0], cost_arr)
        return dist, cost_arr, (px_arr, py_arr)

def dtw_sakoe_chiba(x, y, k, dist_only=True, metric='euclidean'):
    """DTW constrained by Sakoe & Chiba band of width 2k+1.
       The warping path is constrained by |i-j| <= k

    :Parameters:
       x : 1d array_like object (N)
          first sequence
       y : 1d array_like object (M)
          second sequence
       dist_only : bool
          compute only the distance
       metric : 'euclidean', 'sqeuclidean' or 'cosine'
          distance metric to use
    :Returns:
       dist : float
          unnormalized minimum-distance warp path
          between sequences
       cost : 2d numpy array (N,M) [if dist_only=False]
          accumulated cost matrix
       path : tuple of two 1d numpy array (path_x, path_y) [if dist_only=False]
          warp path

     .. [Sakoe78] H Sakoe, & S Chiba S. Dynamic programming algorithm optimization for spoken word recognition. Acoustics, 1978
     """
    return dtw_std(x, y, dist_only=dist_only, metric=metric, constraint='sakoe_chiba', k=k)

def dtw_slanted_band(x, y, k, dist_only=True, metric='euclidean'):
    """DTW constrained by slanted band of width 2k+1.
       The warping path is constrained by |i*len(x)/len(k)-j| <= k.

       Similar to Sakoe & Chiba band constraint, see `dtw_sakoe_chiba`.

    :Parameters:
       x : 1d array_like object (N)
          first sequence
       y : 1d array_like object (M)
          second sequence
       dist_only : bool
          compute only the distance
       metric : 'euclidean', 'sqeuclidean' or 'cosine'
          distance metric to use
    :Returns:
       dist : float
          unnormalized minimum-distance warp path
          between sequences
       cost : 2d numpy array (N,M) [if dist_only=False]
          accumulated cost matrix
       path : tuple of two 1d numpy array (path_x, path_y) [if dist_only=False]
          warp path

     """
    return dtw_std(x, y, dist_only=dist_only, metric=metric, constraint='slanted_band', k=k)


def dtw_itakura(x, y, dist_only=True, metric='euclidean'):
    """DTW constrained by Itakura Parallelogram

    :Parameters:
       x : 1d array_like object (N)
          first sequence
       y : 1d array_like object (M)
          second sequence
       dist_only : bool
          compute only the distance
       metric : 'euclidean', 'sqeuclidean' or 'cosine'
          distance metric to use
    :Returns:
       dist : float
          unnormalized minimum-distance warp path
          between sequences
       cost : 2d numpy array (N,M) [if dist_only=False]
          accumulated cost matrix
       path : tuple of two 1d numpy array (path_x, path_y) [if dist_only=False]
          warp path

    .. [Itakura75] F Itakura. Minimum prediction residual principle applied to speech recognition. Acoustics, Speech and Signal Processing, IEEE Transactions on, 23(1), 67–72, 1975. doi:10.1109/TASSP.1975.1162641.
    """
    return dtw_std(x, y, dist_only=dist_only, metric=metric, constraint='itakura')

def dtw_subsequence(x, y):
    """Subsequence DTW as described in [Muller07]_,
    assuming that the length of `y` is much larger 
    than the length of `x` and using the Manhattan 
    distance (absolute value of the difference) as 
    local cost measure.

    Returns the subsequence of `y` that are close to `x` 
    with respect to the minimum DTW distance.
    
    :Parameters:
       x : 1d array_like object (N)
          first sequence
       y : 1d array_like object (M)
          second sequence

    :Returns:
       dist : float
          unnormalized minimum-distance warp path
          between x and the subsequence of y
       cost : 2d numpy array (N,M) [if dist_only=False]
          complete accumulated cost matrix
       path : tuple of two 1d numpy array (path_x, path_y)
          warp path

    """

    cdef np.ndarray[np.float_t, ndim=1] x_arr
    cdef np.ndarray[np.float_t, ndim=1] y_arr
    cdef np.ndarray[np.float_t, ndim=2] cost_arr
    cdef np.ndarray[np.int_t, ndim=1] px_arr
    cdef np.ndarray[np.int_t, ndim=1] py_arr
    cdef Path p
    cdef int i
    
    x_arr = np.ascontiguousarray(x, dtype=np.float)
    y_arr = np.ascontiguousarray(y, dtype=np.float)
    cost_arr = np.empty((x_arr.shape[0], y_arr.shape[0]), dtype=np.float)

    subsequence(<double *> x_arr.data, <double *> y_arr.data, 
                 <int> x_arr.shape[0], <int> y_arr.shape[0],
                 <double *> cost_arr.data)
    
    idx = np.argmin(cost_arr[-1, :])
    dist = cost_arr[-1, idx]

    subsequence_path(<double *> cost_arr.data, <int> x_arr.shape[0],
                      <int> y_arr.shape[0], <int> idx, &p)
        
    px_arr = np.empty(p.k, dtype=np.int)
    py_arr = np.empty(p.k, dtype=np.int)
    
    for i in range(p.k):
        px_arr[i] = p.px[i]
        py_arr[i] = p.py[i]
            
    free (p.px)
    free (p.py)

    return dist, cost_arr, (px_arr, py_arr)
